#!/usr/bin/perl

use strict;
use warnings;

package VProtect::Restore;
use Carp qw/carp cluck/;
use XML::LibXML;
use Data::Dumper;
use VMware::VIRuntime;
use Socket;
require 'sys/ioctl.ph';

use POSIX qw/strftime/;

use MIME::Base64;

#use VProtect qw/all/;
require 'datastore.pl';

my $devices_hash = {
	'ParaVirtualSCSIController'    => [],
	'VirtualBusLogicController'    => [],
	'VirtualCdrom'                 => [],
	'VirtualDisk'                  => [],
	'VirtualE1000'                 => [],
	'VirtualE1000e'                => [],
	'VirtualFloppy'                => [],
	'VirtualHdAudioCard'           => [],
	'VirtualEnsoniq1371'	       => [],
	'VirtualSoundBlaster16'	       => [],
	'VirtualIDEController'         => [],
	'VirtualLsiLogicController'    => [],
	'VirtualLsiLogicSASController' => [],
	'VirtualAHCIController'	       => [],
	'VirtualMachineVideoCard'      => [],
	'VirtualMachineVMCIDevice'     => [],
	'VirtualMachineVMIROM'         => [],
	'VirtualParallelPort'          => [],
	'VirtualPCIController'         => [],
	'VirtualPCNet32'               => [],
	'VirtualSCSIPassthrough'       => [],
	'VirtualSerialPort'            => [],
	'VirtualUSBController'         => [],
	'VirtualUSBXHCIController'     => [],
	'VirtualVmxnet'                => [],
	'VirtualVmxnet2'               => [],
	'VirtualVmxnet3'               => [],
	'VirtualSriovEthernetCard'     => []
};

my $nas_ds_moref;
my $nas_ds_name;

sub load_xml {
	my $xml_filename = shift;

	my $parser = XML::LibXML->new();

	my $doc = $parser->parse_file($xml_filename);

	return $doc;
}


sub create_vm_config_spec {
	# Create a base configuration spec for a VM that is being restored
	my $opts = shift;

	my $spec = VirtualMachineConfigSpec->new(
		%$opts
	);

	return $spec;
}

sub get_field {
	my $xmlnode = shift;
	my $tag = shift;

	if (not defined $xmlnode) {
		#cluck("not defined!");  # goes to stderr, will end up in vprotect's logs
		return;
	}

	my @results = $xmlnode->getChildrenByTagName($tag);
	if (not defined $results[0]) {
		return;
	}
	return $results[0]->textContent;
}

sub get_field_with_default {
	my $xmlnode = shift;
	my $tag = shift;
	my $default = shift;
	
	if (not defined $xmlnode) {
		return $default;
	}

	my @results = $xmlnode->getChildrenByTagName($tag);
	if (not defined $results[0]) {
		return $default;
	}
	return $results[0]->textContent;
}

sub get_child {
	my $xmlnode = shift;
	my $tag = shift;

	my @results = $xmlnode->getChildrenByTagName($tag);
	if (not defined $results[0]) {
		cluck("couldn't find child node $tag");
		return;
	}
	return $results[0];
}

sub assemble_nic_backing {
	my $device_ref = shift;
	my $device = $$device_ref;

	my $backing;

	my $in_passthrough_mode = $device->{'backing'}{'inPassthroughMode'};
	if (not defined($in_passthrough_mode)) {
		$in_passthrough_mode = 0;
	}

	my $use_autodetect = $device->{'backing'}{'useAutoDetect'};
	if (not defined($use_autodetect)) {
		$use_autodetect = 0;
	}

	my $device_name = $device->{'backing'}{'deviceName'};
	if (not defined($device_name)) {
		$device_name = "Unknown device";
	}

	# Always restore as a standard Network backing.
	# Restoring a dvSwitch backing is not especially useful since the network may
	# not exist at restore time, and if it doesn't, the result for the user is the
	# same as if we restored a standard network backing.  (Requires choosing
	# network before starting the VM.)

	return VirtualEthernetCardNetworkBackingInfo->new(
		deviceName => $device_name,
		inPassthroughMode => $in_passthrough_mode,
		useAutoDetect => $use_autodetect
	);
}

sub process_device {
	# Identify the device and load the fields we need from the XML.

	# We will later use this information to build a configspec for the VM.
	# This should mirror the get_vm_information_by_moref used for backup.

	my $dx = shift;  # device xml node

	my $name = $dx->getName;

	if ($name eq 'ParaVirtualSCSIController') {
		my $deviceInfo_x = get_child($dx, 'deviceInfo');
		
		my $devref = {};
		$devref->{'busNumber'}          = get_field($dx, 'busNumber');
		$devref->{'controllerKey'}      = get_field($dx, 'controllerKey');
		$devref->{'hotAddRemove'}       = get_field($dx, 'hotAddRemove');
		$devref->{'key'}                = get_field($dx, 'key');
		$devref->{'scsiCtlrUnitNumber'} = get_field($dx, 'scsiCtlrUnitNumber');
		$devref->{'sharedBus'}          = get_field($dx, 'sharedBus');
		$devref->{'unitNumber'}         = get_field($dx, 'unitNumber');

		if (defined $deviceInfo_x) {
			$devref->{'deviceInfo'}{'label'}   = get_field($deviceInfo_x, 'label');
			$devref->{'deviceInfo'}{'summary'} = get_field($deviceInfo_x, 'summary');
		}
		
		my $elem = $devices_hash->{'ParaVirtualSCSIController'};
		push (@$elem, $devref);
	}
	elsif ($name eq 'VirtualBusLogicController') {
		my $deviceInfo_x = get_child($dx, 'deviceInfo');

		my $devref = {};
		$devref->{'busNumber'}          = get_field($dx, 'busNumber');
		$devref->{'controllerKey'}      = get_field($dx, 'controllerKey');
		$devref->{'hotAddRemove'}       = get_field($dx, 'hotAddRemove');
		$devref->{'key'}                = get_field($dx, 'key');
		$devref->{'scsiCtlrUnitNumber'} = get_field($dx, 'scsiCtlrUnitNumber');
		$devref->{'unitNumber'}         = get_field($dx, 'unitNumber');

		my $elem = $devices_hash->{'VirtualBusLogicController'};
		push (@$elem, $devref);
	}
	elsif ($name eq 'VirtualCdrom') {
		my $backing_x     = get_child($dx, 'backing');
		my $connectable_x = get_child($dx, 'connectable');
		my $deviceInfo_x  = get_child($dx, 'deviceInfo');

		my $devref = {};
		$devref->{'controllerKey'} = get_field($dx, 'controllerKey');
		$devref->{'key'}           = get_field($dx, 'key');
		$devref->{'unitNumber'}    = get_field($dx, 'unitNumber');
		$devref->{'deviceInfo'}{'label'}   = get_field($deviceInfo_x, 'label');
		$devref->{'deviceInfo'}{'summary'} = get_field($deviceInfo_x, 'summary');

		$devref->{'backing'}{'_backing_type'} = get_field($backing_x, '_backing_type');
		$devref->{'backing'}{'deviceName'}    = get_field($backing_x, 'deviceName');
		$devref->{'backing'}{'exclusive'}     = get_field($backing_x, 'exclusive');
		$devref->{'backing'}{'fileName'}      = get_field($backing_x, 'fileName');
		$devref->{'backing'}{'useAutoDetect'} = get_field($backing_x, 'useAutoDetect');

		$devref->{'connectable'}{'allowGuestControl'} = get_field($connectable_x, 'allowGuestControl');
		$devref->{'connectable'}{'connected'}         = get_field($connectable_x, 'connected');
		$devref->{'connectable'}{'startConnected'}    = get_field($connectable_x, 'startConnected');
		$devref->{'connectable'}{'status'}            = get_field($connectable_x, 'status');

		my $elem = $devices_hash->{'VirtualCdrom'};
		push @$elem, $devref;
	}
	elsif ($name eq 'VirtualDisk') {
		my $backing_x = get_child($dx, 'backing');
		my $deviceInfo_x = get_child($dx, 'deviceInfo');
		
		my $devref = {};
		$devref->{'capacityInKB'}  = get_field($dx, 'capacityInKB');
		$devref->{'controllerKey'} = get_field($dx, 'controllerKey');
		$devref->{'key'}           = get_field($dx, 'key');
		# Bug 16541 - Add this information to know later which disks were excluded
		$devref->{'is_excluded'}   = get_field($dx, 'is_excluded');
		$devref->{'unitNumber'}    = get_field($dx, 'unitNumber');

		$devref->{'backing'}{'_backing_type'}   = get_field($backing_x, '_backing_type');
		$devref->{'backing'}{'diskMode'}        = get_field($backing_x, 'diskMode');
		$devref->{'backing'}{'fileName'}        = "";
		$devref->{'backing'}{'thinProvisioned'} = get_field($backing_x, 'thinProvisioned');
		$devref->{'backing'}{'uuid'}            = get_field($backing_x, 'uuid');
		$devref->{'backing'}{'writeThrough'}    = get_field($backing_x, 'writeThrough');

		$devref->{'deviceInfo'}{'label'}   = get_field($deviceInfo_x, 'label');
		$devref->{'deviceInfo'}{'summary'} = get_field($deviceInfo_x, 'summary');

		my $elem = $devices_hash->{'VirtualDisk'};
		push (@$elem, $devref);
	}
	elsif ($name eq 'VirtualE1000' or $name eq 'VirtualE1000e') {
		my $backing_x     = get_child($dx, 'backing');
		my $connectable_x = get_child($dx, 'connectable');
		my $deviceInfo_x  = get_child($dx, 'deviceInfo');

		my $devref = {};
		$devref->{'addressType'}      = get_field($dx, 'addressType');
		$devref->{'controllerKey'}    = get_field($dx, 'controllerKey');
		$devref->{'key'}              = get_field($dx, 'key');
		$devref->{'macAddress'}       = get_field($dx, 'macAddress');
		$devref->{'unitNumber'}       = get_field($dx, 'unitNumber');
		$devref->{'wakeOnLanEnabled'} = get_field($dx, 'wakeOnLanEnabled');
		$devref->{'deviceInfo'}{'label'}   = get_field($deviceInfo_x, 'label');
		$devref->{'deviceInfo'}{'summary'} = get_field($deviceInfo_x, 'summary');
		$devref->{'backing'}{'_backing_type'}     = get_field($backing_x, '_backing_type');
		$devref->{'backing'}{'deviceName'}        = get_field($backing_x, 'deviceName');
		$devref->{'backing'}{'inPassthroughMode'} = get_field($backing_x, 'inPassthroughMode');
		$devref->{'backing'}{'useAutoDetect'}     = get_field($backing_x, 'useAutoDetect');
		$devref->{'backing'}{'switchUuid'}        = get_field($backing_x, 'switchUuid');
		$devref->{'connectable'}{'allowGuestControl'} = get_field($connectable_x, 'allowGuestControl');
		$devref->{'connectable'}{'connected'}         = get_field($connectable_x, 'connected');
		$devref->{'connectable'}{'startConnected'}    = get_field($connectable_x, 'startConnected');

		my $elem = $devices_hash->{$name};
		push @$elem, $devref;
	}
	elsif ($name eq 'VirtualFloppy') {
		my $backing_x     = get_child($dx, 'backing');
		my $connectable_x = get_child($dx, 'connectable');
		my $deviceInfo_x  = get_child($dx, 'deviceInfo');

		my $devref = {};
		$devref->{'controllerKey'} = get_field($dx, 'controllerKey');
		$devref->{'key'}           = get_field($dx, 'key');
		$devref->{'unitNumber'}    = get_field($dx, 'unitNumber');
		$devref->{'backing'}{'_backing_type'} = get_field($backing_x, '_backing_type');
		$devref->{'backing'}{'deviceName'}    = get_field($backing_x, 'deviceName');
		$devref->{'backing'}{'fileName'}      = get_field($backing_x, 'fileName');
		$devref->{'backing'}{'useAutoDetect'} = get_field($backing_x, 'useAutoDetect');
		$devref->{'deviceInfo'}{'label'}   = get_field($deviceInfo_x, 'label');
		$devref->{'deviceInfo'}{'summary'} = get_field($deviceInfo_x, 'summary');
		$devref->{'connectable'}{'allowGuestControl'} = get_field($connectable_x, 'allowGuestControl');
		$devref->{'connectable'}{'connected'}         = get_field($connectable_x, 'connected');
		$devref->{'connectable'}{'startConnected'}    = get_field($connectable_x, 'startConnected');

		my $elem = $devices_hash->{'VirtualFloppy'};
		push @$elem, $devref;
	}
	elsif ($name eq 'VirtualHdAudioCard') {
		my $backing_x     = get_child($dx, 'backing');
		my $connectable_x = get_child($dx, 'connectable');
		my $deviceInfo_x  = get_child($dx, 'deviceInfo');

		my $devref = {};
		$devref->{'key'} = get_field($dx, 'key');
		$devref->{'backing'}{'_backing_type'} = get_field($backing_x, '_backing_type');
		$devref->{'backing'}{'deviceName'}    = get_field($backing_x, 'deviceName');
		$devref->{'backing'}{'useAutoDetect'} = get_field_with_default($backing_x, 'useAutoDetect', 1);
		$devref->{'deviceInfo'}{'label'}   = get_field($deviceInfo_x, 'label');
		$devref->{'deviceInfo'}{'summary'} = get_field($deviceInfo_x, 'summary');
		$devref->{'connectable'}{'allowGuestControl'} = get_field_with_default($connectable_x, 'allowGuestControl', 0);
		$devref->{'connectable'}{'connected'}         = get_field_with_default($connectable_x, 'connected', 0);
		$devref->{'connectable'}{'startConnected'}    = get_field_with_default($connectable_x, 'startConnected', 1);

		my $elem = $devices_hash->{'VirtualHdAudioCard'};
		push @$elem, $devref;
	}
	elsif ($name eq 'VirtualEnsoniq1371') {
		my $backing_x     = get_child($dx, 'backing');
		my $connectable_x = get_child($dx, 'connectable');
		my $deviceInfo_x  = get_child($dx, 'deviceInfo');

		my $devref = {};
		$devref->{'key'} = get_field($dx, 'key');
		$devref->{'backing'}{'_backing_type'} = get_field($backing_x, '_backing_type');
		$devref->{'backing'}{'deviceName'}    = get_field($backing_x, 'deviceName');
		$devref->{'backing'}{'useAutoDetect'} = get_field_with_default($backing_x, 'useAutoDetect', 1);
		$devref->{'deviceInfo'}{'label'}   = get_field($deviceInfo_x, 'label');
		$devref->{'deviceInfo'}{'summary'} = get_field($deviceInfo_x, 'summary');
		$devref->{'connectable'}{'allowGuestControl'} = get_field_with_default($connectable_x, 'allowGuestControl', 0);
		$devref->{'connectable'}{'connected'}         = get_field_with_default($connectable_x, 'connected', 0);
		$devref->{'connectable'}{'startConnected'}    = get_field_with_default($connectable_x, 'startConnected', 1);

		my $elem = $devices_hash->{'VirtualEnsoniq1371'};
		push @$elem, $devref;
	}
	elsif ($name eq 'VirtualSoundBlaster16') {
		my $backing_x     = get_child($dx, 'backing');
		my $connectable_x = get_child($dx, 'connectable');
		my $deviceInfo_x  = get_child($dx, 'deviceInfo');

		my $devref = {};
		$devref->{'key'} = get_field($dx, 'key');
		$devref->{'backing'}{'_backing_type'} = get_field($backing_x, '_backing_type');
		$devref->{'backing'}{'deviceName'}    = get_field($backing_x, 'deviceName');
		$devref->{'backing'}{'useAutoDetect'} = get_field_with_default($backing_x, 'useAutoDetect', 1);
		$devref->{'deviceInfo'}{'label'}   = get_field($deviceInfo_x, 'label');
		$devref->{'deviceInfo'}{'summary'} = get_field($deviceInfo_x, 'summary');
		$devref->{'connectable'}{'allowGuestControl'} = get_field_with_default($connectable_x, 'allowGuestControl', 0);
		$devref->{'connectable'}{'connected'}         = get_field_with_default($connectable_x, 'connected', 0);
		$devref->{'connectable'}{'startConnected'}    = get_field_with_default($connectable_x, 'startConnected', 1);

		my $elem = $devices_hash->{'VirtualSoundBlaster16'};
		push @$elem, $devref;
	}
	elsif ($name eq 'VirtualKeyboard') {
		# This should not appear in a backup.  If it somehow does, ignore it.
	}
	elsif ($name eq 'VirtualIDEController') {
		# TODO: this is on the list of things not to restore
		my $devref = {};
		my $deviceInfo_x = get_child($dx, 'deviceInfo');
		$devref->{'busNumber'} = get_field($dx, 'busNumber');
		$devref->{'key'} = get_field($dx, 'key');

		if (defined $deviceInfo_x) {
			$devref->{'deviceInfo'}{'label'}   = get_field($deviceInfo_x, 'label');
			$devref->{'deviceInfo'}{'summary'} = get_field($deviceInfo_x, 'summary');
		}

		my $elem = $devices_hash->{'VirtualIDEController'};
		push @$elem, $devref;
	}
	elsif ($name eq 'VirtualLsiLogicController') {
		my $devref = {};

		$devref->{'busNumber'}          = get_field($dx, 'busNumber');
		$devref->{'controllerKey'}      = get_field($dx, 'controllerKey');
		$devref->{'hotAddRemove'}       = get_field($dx, 'hotAddRemove');
		$devref->{'key'}                = get_field($dx, 'key');
		$devref->{'scsiCtlrUnitNumber'} = get_field($dx, 'scsiCtlrUnitNumber');
		$devref->{'sharedBus'}          = get_field($dx, 'sharedBus');
		$devref->{'unitNumber'}         = get_field($dx, 'unitNumber');

		my $elem = $devices_hash->{'VirtualLsiLogicController'};
		push @$elem, $devref;
	}
	elsif ($name eq 'VirtualLsiLogicSASController') {
		my $devref = {};

		$devref->{'busNumber'}          = get_field($dx, 'busNumber');
		$devref->{'controllerKey'}      = get_field($dx, 'controllerKey');
		$devref->{'hotAddRemove'}       = get_field($dx, 'hotAddRemove');
		$devref->{'key'}                = get_field($dx, 'key');
		$devref->{'scsiCtlrUnitNumber'} = get_field($dx, 'scsiCtlrUnitNumber');
		$devref->{'sharedBus'}          = get_field($dx, 'sharedBus');
		$devref->{'unitNumber'}         = get_field($dx, 'unitNumber');

		my $elem = $devices_hash->{'VirtualLsiLogicSASController'};
		push @$elem, $devref;

	}
	elsif ($name eq 'VirtualAHCIController') {
		# Supported from vSphere 5.5 and Unitrends Version 7.4
		my $devref = {};

		$devref->{'busNumber'}          = get_field($dx, 'busNumber');
		$devref->{'controllerKey'}      = get_field($dx, 'controllerKey');
		$devref->{'key'}                = get_field($dx, 'key');
		$devref->{'unitNumber'}         = get_field($dx, 'unitNumber');

		my $elem = $devices_hash->{'VirtualAHCIController'};
		push @$elem, $devref;

	}
	elsif ($name eq 'VirtualMachineVideoCard') {
		# This should not appear in a backup.  If it somehow does, ignore it.
		my $devref = {};

		$devref->{'controllerKey'}    = get_field($dx, 'controllerKey');
		$devref->{'key'}              = get_field($dx, 'key');
		$devref->{'unitNumber'}       = get_field($dx, 'unitNumber');

		$devref->{'enable3DSupport'}  = get_field($dx, 'enable3DSupport');
		$devref->{'numDisplays'}      = get_field($dx, 'numDisplays');
		$devref->{'useAutoDetect'}    = get_field($dx, 'useAutoDetect');
		$devref->{'videoRamSizeInKB'} = get_field($dx, 'videoRamSizeInKB');

		#my $elem = $devices_hash->{'VirtualMachineVideoCard'};
		#push @$elem, $devref;
	}
	elsif ($name eq 'VirtualMachineVMCIDevice') {
		my $deviceInfo_x = get_child($dx, 'deviceInfo');

		my $devref = {};
		$devref->{'allowUnrestrictedCommunication'} = get_field($dx, 'allowUnrestrictedCommunication');
		$devref->{'controllerKey'}                  = get_field($dx, 'controllerKey');
		$devref->{'key'}                            = get_field($dx, 'key');
		$devref->{'unitNumber'}                     = get_field($dx, 'unitNumber');
		$devref->{'deviceInfo'}{'label'}            = get_field($deviceInfo_x, 'label');
		$devref->{'deviceInfo'}{'summary'}          = get_field($deviceInfo_x, 'summary');

		my $elem = $devices_hash->{'VirtualMachineVMCIDevice'};
		push @$elem, $devref;
	}
	elsif ($name eq 'VirtualMachineVMIROM') {
		my $devref = {};
		$devref->{'present'}  = 1;

		my $elem = $devices_hash->{'VirtualMachineVMIROM'};
		push @$elem, $devref;
	}
	elsif ($name eq 'VirtualParallelPort') {
		my $backing_x = get_child($dx, 'backing');
		my $connectable_x = get_child($dx, 'connectable');
		my $deviceInfo_x = get_child($dx, 'deviceInfo');

		my $devref = {};
		$devref->{'controllerKey'}   = get_field($dx, 'controllerKey');
		$devref->{'key'}             = get_field($dx, 'key');
		$devref->{'unitNumber'}      = get_field($dx, 'unitNumber');
		$devref->{'backing'}{'_backing_type'} = get_field($backing_x, '_backing_type');
		$devref->{'backing'}{'deviceName'}    = get_field($backing_x, 'deviceName');
		$devref->{'backing'}{'fileName'}      = get_field($backing_x, 'fileName');
		$devref->{'backing'}{'useAutoDetect'} = get_field($backing_x, 'useAutoDetect');
		$devref->{'connectable'}{'allowGuestControl'} = get_field($connectable_x, 'allowGuestControl');
		$devref->{'connectable'}{'connected'}         = get_field($connectable_x, 'connected');
		$devref->{'connectable'}{'start'}             = get_field($connectable_x, 'start');
		$devref->{'connectable'}{'startConnected'}    = get_field($connectable_x, 'startConnected');
		$devref->{'connectable'}{'status'}            = get_field($connectable_x, 'status');
		$devref->{'deviceInfo'}{'label'}   = get_field($deviceInfo_x, 'deviceInfo');
		$devref->{'deviceInfo'}{'summary'} = get_field($deviceInfo_x, 'summary');

		my $elem = $devices_hash->{'VirtualParallelPort'};
		push @$elem, $devref;
	}
	elsif ($name eq 'VirtualPCIController') {
		my $deviceInfo_x = get_child($dx, 'deviceInfo');

		my $devref = {};
		$devref->{'busNumber'} = get_field($dx, 'busNumber');
		$devref->{'key'} =       get_field($dx, 'key');
		$devref->{'deviceInfo'}{'label'}   = get_field($deviceInfo_x, 'label');
		$devref->{'deviceInfo'}{'summary'} = get_field($deviceInfo_x, 'summary');

		my $elem = $devices_hash->{'VirtualPCIController'};
		push @$elem, $devref;
	}
	elsif ($name eq 'VirtualPCNet32') {
		my $backing_x     = get_child($dx, 'backing');
		my $connectable_x = get_child($dx, 'connectable');
		my $deviceInfo_x  = get_child($dx, 'deviceInfo');

		my $devref = {};
		$devref->{'addressType'}      = get_field($dx, 'addressType');
		$devref->{'controllerKey'}    = get_field($dx, 'controllerKey');
		$devref->{'key'}              = get_field($dx, 'key');
		$devref->{'macAddress'}       = get_field($dx, 'macAddress');
		$devref->{'unitNumber'}       = get_field($dx, 'unitNumber');
		$devref->{'wakeOnLanEnabled'} = get_field($dx, 'wakeOnLanEnabled');
		$devref->{'backing'}{'_backing_type'}     = get_field($backing_x, '_backing_type');
		$devref->{'backing'}{'deviceName'}        = get_field($backing_x, 'deviceName');
		$devref->{'backing'}{'inPassthroughMode'} = get_field($backing_x, 'inPassthroughMode');
		$devref->{'backing'}{'useAutoDetect'}     = get_field($backing_x, 'useAutoDetect');
		$devref->{'backing'}{'switchUuid'}        = get_field($backing_x, 'switchUuid');
		$devref->{'connectable'}{'allowGuestControl'} = get_field($connectable_x, 'allowGuestControl');
		$devref->{'connectable'}{'connected'}         = get_field($connectable_x, 'connected');
		$devref->{'connectable'}{'startConnected'}    = get_field($connectable_x, 'startConnected');
		$devref->{'deviceInfo'}{'label'}   = get_field($deviceInfo_x, 'label');
		$devref->{'deviceInfo'}{'summary'} = get_field($deviceInfo_x, 'summary');

		my $elem = $devices_hash->{'VirtualPCNet32'};
		push @$elem, $devref;
	}
	elsif ($name eq 'VirtualPS2Controller') {
		my $deviceInfo_x = get_child($dx, 'deviceInfo');
		my $devref = {};
		$devref->{'key'} = get_field($dx, 'key');
		$devref->{'deviceInfo'}{'label'}   = get_field($deviceInfo_x, 'label');
		$devref->{'deviceInfo'}{'summary'} = get_field($deviceInfo_x, 'summary');
	
		my $elem = $devices_hash->{'VirtualPS2Controller'};
		push @$elem, $devref;
	}
	elsif ($name eq 'VirtualPointingDevice') {
		# This should not appear in a backup.  If it somehow does, ignore it.
	}
	elsif ($name eq 'VirtualSCSIPassthrough') {
		my $backing_x     = get_child($dx, 'backing');
		my $connectable_x = get_child($dx, 'connectable');
		my $deviceInfo_x  = get_child($dx, 'deviceInfo');

		my $devref = {};
		$devref->{'controllerKey'}                    = get_field($dx, 'controllerKey');
		$devref->{'key'}                              = get_field($dx, 'key');
		$devref->{'unitNumber'}                       = get_field($dx, 'unitNumber');
		$devref->{'backing'}{'_backing_type'}         = get_field($backing_x, '_backing_type');
		$devref->{'backing'}{'deviceName'}            = get_field($backing_x, 'deviceName');
		$devref->{'backing'}{'useAutoDetect'}         = get_field($backing_x, 'useAutoDetect');
		$devref->{'connectable'}{'allowGuestControl'} = get_field($connectable_x, 'allowGuestControl');
		$devref->{'connectable'}{'connected'}         = get_field($connectable_x, 'connected');
		$devref->{'connectable'}{'startConnected'}    = get_field($connectable_x, 'startConnected');
		$devref->{'deviceInfo'}{'label'}              = get_field($deviceInfo_x, 'label');
		$devref->{'deviceInfo'}{'summary'}            = get_field($deviceInfo_x, 'summary');

		my $elem = $devices_hash->{'VirtualSCSIPassthrough'};
		push @$elem, $devref;
	}
	elsif ($name eq 'VirtualSerialPort') {
		my $backing_x = get_child($dx, 'backing');
		my $connectable_x = get_child($dx, 'connectable');
		my $deviceInfo_x = get_child($dx, 'deviceInfo');

		my $devref = {};
		$devref->{'controllerKey'}   = get_field($dx, 'controllerKey');
		$devref->{'key'}             = get_field($dx, 'key');
		$devref->{'unitNumber'}      = get_field($dx, 'unitNumber');
		$devref->{'yieldOnPoll'}     = get_field($dx, 'yieldOnPoll');
		$devref->{'backing'}{'_backing_type'} = get_field($backing_x, '_backing_type');
		$devref->{'backing'}{'deviceName'}    = get_field($backing_x, 'deviceName');
		$devref->{'backing'}{'direction'}     = get_field($backing_x, 'direction');
		$devref->{'backing'}{'endpoint'}      = get_field($backing_x, 'endpoint');
		$devref->{'backing'}{'fileName'}      = get_field($backing_x, 'fileName');
		$devref->{'backing'}{'noRxLoss'}      = get_field($backing_x, 'noRxLoss');
		$devref->{'backing'}{'pipeName'}      = get_field($backing_x, 'pipeName');
		$devref->{'backing'}{'proxyURI'}      = get_field($backing_x, 'proxyURI');
		$devref->{'backing'}{'serviceURI'}    = get_field($backing_x, 'serviceURI');
		$devref->{'backing'}{'useAutoDetect'} = get_field($backing_x, 'useAutoDetect');
		$devref->{'connectable'}{'allowGuestControl'} = get_field($connectable_x, 'allowGuestControl');
		$devref->{'connectable'}{'connected'}         = get_field($connectable_x, 'connected');
		$devref->{'connectable'}{'startConnected'}    = get_field($connectable_x, 'startConnected');
		$devref->{'connectable'}{'status'}            = get_field($connectable_x, 'status');
		$devref->{'deviceInfo'}{'label'}   = get_field($deviceInfo_x, 'label');
		$devref->{'deviceInfo'}{'summary'} = get_field($deviceInfo_x, 'summary');

		my $elem = $devices_hash->{'VirtualSerialPort'};
		push @$elem, $devref;
	}
	elsif ($name eq 'VirtualSIOController') {
		my $deviceInfo_x = get_child($dx, 'deviceInfo');
		my $devref = {};
		$devref->{'key'} = get_field($dx, 'key');
		$devref->{'deviceInfo'}{'label'}   = get_field($deviceInfo_x, 'label');
		$devref->{'deviceInfo'}{'summary'} = get_field($deviceInfo_x, 'summary');
	
		my $elem = $devices_hash->{'VirtualSIOController'};
		push @$elem, $devref;
	}
	elsif ($name eq 'VirtualUSBController' or $name eq 'VirtualUSBXHCIController') {
		my $devref = {};

		$devref->{'autoConnectDevices'} = get_field_with_default($dx, 'autoConnectDevices', 0);
		$devref->{'busNumber'}          = get_field($dx, 'busNumber');
		$devref->{'ehciEnabled'}        = get_field($dx, 'ehciEnabled');
		$devref->{'key'}                = get_field($dx, 'key');
		$devref->{'unitNumber'}         = get_field($dx, 'unitNumber');

		my $elem = $devices_hash->{$name};
		push @$elem, $devref;
	}
	elsif ($name eq 'VirtualVmxnet') {
		my $backing_x = get_child($dx, 'backing');
		my $connectable_x = get_child($dx, 'connectable');
		my $deviceInfo_x = get_child($dx, 'deviceInfo');

		my $devref = {};
		$devref->{'addressType'}      = get_field($dx, 'addressType');
		$devref->{'controllerKey'}    = get_field($dx, 'controllerKey');
		$devref->{'key'}              = get_field($dx, 'key');
		$devref->{'macAddress'}       = get_field($dx, 'macAddress');
		$devref->{'unitNumber'}       = get_field($dx, 'unitNumber');
		$devref->{'wakeOnLanEnabled'} = get_field($dx, 'wakeOnLanEnabled');
		$devref->{'backing'}{'_backing_type'}     = get_field($backing_x, '_backing_type');
		$devref->{'backing'}{'deviceName'}        = get_field($backing_x, 'deviceName');
		$devref->{'backing'}{'inPassthroughMode'} = get_field($backing_x, 'inPassthroughMode');
		$devref->{'backing'}{'useAutoDetect'}     = get_field($backing_x, 'useAutoDetect');
		$devref->{'backing'}{'switchUuid'}        = get_field($backing_x, 'switchUuid');
		$devref->{'connectable'}{'allowGuestControl'} = get_field($connectable_x, 'allowGuestControl');
		$devref->{'connectable'}{'connected'}         = get_field($connectable_x, 'connected');
		$devref->{'connectable'}{'startConnected'}    = get_field($connectable_x, 'startConnected');
		$devref->{'deviceInfo'}{'label'}   = get_field($deviceInfo_x, 'label');
		$devref->{'deviceInfo'}{'summary'} = get_field($deviceInfo_x, 'summary');

		my $elem = $devices_hash->{'VirtualVmxnet'};
		push @$elem, $devref;
	}
	elsif ($name eq 'VirtualVmxnet2') {
		my $backing_x = get_child($dx, 'backing');
		my $connectable_x = get_child($dx, 'connectable');
		my $deviceInfo_x = get_child($dx, 'deviceInfo');

		my $devref = {};
		$devref->{'addressType'}      = get_field($dx, 'addressType');
		$devref->{'controllerKey'}    = get_field($dx, 'controllerKey');
		$devref->{'key'}              = get_field($dx, 'key');
		$devref->{'macAddress'}       = get_field($dx, 'macAddress');
		$devref->{'unitNumber'}       = get_field($dx, 'unitNumber');
		$devref->{'wakeOnLanEnabled'} = get_field($dx, 'wakeOnLanEnabled');
		$devref->{'backing'}{'_backing_type'}     = get_field($backing_x, '_backing_type');
		$devref->{'backing'}{'deviceName'}        = get_field($backing_x, 'deviceName');
		$devref->{'backing'}{'inPassthroughMode'} = get_field($backing_x, 'inPassthroughMode');
		$devref->{'backing'}{'useAutoDetect'}     = get_field($backing_x, 'useAutoDetect');
		$devref->{'backing'}{'switchUuid'}        = get_field($backing_x, 'switchUuid');
		$devref->{'connectable'}{'allowGuestControl'} = get_field($connectable_x, 'allowGuestControl');
		$devref->{'connectable'}{'connected'}         = get_field($connectable_x, 'connected');
		$devref->{'connectable'}{'startConnected'}    = get_field($connectable_x, 'startConnected');
		$devref->{'deviceInfo'}{'label'}   = get_field($deviceInfo_x, 'label');
		$devref->{'deviceInfo'}{'summary'} = get_field($deviceInfo_x, 'summary');

		my $elem = $devices_hash->{'VirtualVmxnet2'};
		push @$elem, $devref;
	}
	elsif ($name eq 'VirtualVmxnet3') {
		my $backing_x = get_child($dx, 'backing');
		my $connectable_x = get_child($dx, 'connectable');
		my $deviceInfo_x = get_child($dx, 'deviceInfo');

		my $devref = {};
		$devref->{'addressType'}      = get_field($dx, 'addressType');
		$devref->{'controllerKey'}    = get_field($dx, 'controllerKey');
		$devref->{'macAddress'}       = get_field($dx, 'macAddress');
		$devref->{'key'}              = get_field($dx, 'key');
		$devref->{'unitNumber'}       = get_field($dx, 'unitNumber');
		$devref->{'wakeOnLanEnabled'} = get_field($dx, 'wakeOnLanEnabled');
		$devref->{'backing'}{'_backing_type'}     = get_field($backing_x, '_backing_type');
		$devref->{'backing'}{'deviceName'}        = get_field($backing_x, 'deviceName');
		$devref->{'backing'}{'inPassthroughMode'} = get_field($backing_x, 'inPassthroughMode');
		$devref->{'backing'}{'useAutoDetect'}     = get_field($backing_x, 'useAutoDetect');
		$devref->{'backing'}{'switchUuid'}        = get_field($backing_x, 'switchUuid');
		$devref->{'connectable'}{'allowGuestControl'} = get_field($connectable_x, 'allowGuestControl');
		$devref->{'connectable'}{'connected'}         = get_field($connectable_x, 'connected');
		$devref->{'connectable'}{'startConnected'}    = get_field($connectable_x, 'startConnected');
		$devref->{'deviceInfo'}{'label'}   = get_field($deviceInfo_x, 'label');
		$devref->{'deviceInfo'}{'summary'} = get_field($deviceInfo_x, 'summary');

		my $elem = $devices_hash->{'VirtualVmxnet3'};
		push @$elem, $devref;

	}
	elsif ($name eq 'VirtualSriovEthernetCard') {
		my $backing_x = get_child($dx, 'backing');
		my $connectable_x = get_child($dx, 'connectable');
		my $deviceInfo_x = get_child($dx, 'deviceInfo');

		my $devref = {};
		$devref->{'addressType'}      = get_field($dx, 'addressType');
		$devref->{'controllerKey'}    = get_field($dx, 'controllerKey');
		$devref->{'macAddress'}       = get_field($dx, 'macAddress');
		$devref->{'key'}              = get_field($dx, 'key');
		$devref->{'unitNumber'}       = get_field($dx, 'unitNumber');
		$devref->{'wakeOnLanEnabled'} = get_field($dx, 'wakeOnLanEnabled');
		$devref->{'backing'}{'_backing_type'}     = get_field($backing_x, '_backing_type');
		$devref->{'backing'}{'deviceName'}        = get_field($backing_x, 'deviceName');
		$devref->{'backing'}{'inPassthroughMode'} = get_field($backing_x, 'inPassthroughMode');
		$devref->{'backing'}{'useAutoDetect'}     = get_field($backing_x, 'useAutoDetect');
		$devref->{'connectable'}{'allowGuestControl'} = get_field($connectable_x, 'allowGuestControl');
		$devref->{'connectable'}{'connected'}         = get_field($connectable_x, 'connected');
		$devref->{'connectable'}{'startConnected'}    = get_field($connectable_x, 'startConnected');
		$devref->{'deviceInfo'}{'label'}   = get_field($deviceInfo_x, 'label');
		$devref->{'deviceInfo'}{'summary'} = get_field($deviceInfo_x, 'summary');

		my $elem = $devices_hash->{'VirtualSriovEthernetCard'};
		push @$elem, $devref;

	}
	else {
		die "Unable to identify hardware device of type " . $name . "\n";
	}
}

sub create_device_change_spec {
	# Create a config spec based on the devices hash.
	# This takes the device information generated by reading the XML and creates VMware objects
	#  for the VM configuration.

	# This must be done in order of VM hardware dependencies, i.e. disk controllers before disks, etc.

	my $device_ref = shift;
	my $datastore_moref = shift;   # Datastore disks are created on
	my $datastore_path = shift;    # Datastore disks are created on
	my $metadata = shift;
	my $wir_mode = shift;
	my $annotation_ref = shift;
	my $disk_map_href = shift;     # Only defined if we are doing instant recovery.
	my $restore_dir = shift;       # Only defined if we are doing instant recovery.

	my $instant_mode = 0;
	if (defined ($disk_map_href)) {
		$instant_mode = 1;
	}

	my @device_config_spec;

	sub newspec {
		# Wrap a device object with a configspec operation to create it
		my $dev = shift;
		return VirtualDeviceConfigSpec->new(
				device => $dev,
				operation => VirtualDeviceConfigSpecOperation->new('add')
		);	
	}

	sub newdiskspec {
		# Wrap a disk object with a configspec operation to create it
		my $dev = shift;
		return VirtualDeviceConfigSpec->new(
				device => $dev,
				operation => VirtualDeviceConfigSpecOperation->new('add'),
				fileOperation => VirtualDeviceConfigSpecFileOperation->new('create')
		);
	}

	#print STDERR Dumper($device_ref);

	if (defined $device_ref->{'VirtualPCIController'}) {
		my $devices = $device_ref->{'VirtualPCIController'};

		foreach (@$devices) {
			# Create VirtualPCIController devices
			my $dev = VirtualPCIController->new(
					busNumber =>      $_->{'busNumber'},
					key       => -1 * $_->{'key'},
			);

			push @device_config_spec, newspec($dev);
		}
	}

	
	###################
	# disk controllers
	###################

	if (defined $device_ref->{'VirtualIDEController'}) {
		my $devices = $device_ref->{'VirtualIDEController'};

		foreach (@$devices) {
			# Create VirtualIDEController devices
			#my $dev = VirtualIDEController->new(
			#		busNumber =>      $_->{'busNumber'},
			#		key       => -1 * $_->{'key'});

			#push @device_config_spec, newspec($dev);
		}
	}

	if (defined $device_ref->{'ParaVirtualSCSIController'}) {
		my $devices = $device_ref->{'ParaVirtualSCSIController'};

		foreach (@$devices) {
			# Create ParaVirtualSCSIController
			my $dev = ParaVirtualSCSIController->new(
				busNumber          => $_->{'busNumber'},
				controllerKey      => $_->{'controllerKey'},
				hotAddRemove       => $_->{'hotAddRemove'},
				key                => $_->{'key'},
				scsiCtlrUnitNumber => $_->{'scsiCtlrUnitNumber'},
				sharedBus          => VirtualSCSISharing->new('noSharing'),
				unitNumber         => $_->{'unitNumber'}
			);

			push @device_config_spec, newspec($dev);
		}
	}

	if (defined $device_ref->{'VirtualBusLogicController'}) {
		my $devices = $device_ref->{'VirtualBusLogicController'};

		foreach (@$devices) {
			# Create VirtualBusLogicController
			my $dev = VirtualBusLogicController->new(
				busNumber          => $_->{'busNumber'},
				controllerKey      => $_->{'controllerKey'},
				hotAddRemove       => $_->{'hotAddRemove'},
				key                => $_->{'key'},
				scsiCtlrUnitNumber => $_->{'scsiCtlrUnitNumber'},
				sharedBus          => VirtualSCSISharing->new('noSharing'),
				unitNumber         => $_->{'unitNumber'}
			);

			push @device_config_spec, newspec($dev);
		}
	}

	if (defined $device_ref->{'VirtualLsiLogicController'}) {
		my $devices = $device_ref->{'VirtualLsiLogicController'};

		foreach (@$devices) {
			# Create VirtualLsiLogicController
			my $dev = VirtualLsiLogicController->new(
				busNumber          => $_->{'busNumber'},
				controllerKey      => $_->{'controllerKey'},
				hotAddRemove       => $_->{'hotAddRemove'},
				key                => $_->{'key'},
				scsiCtlrUnitNumber => $_->{'scsiCtlrUnitNumber'},
				sharedBus          => VirtualSCSISharing->new('noSharing'),
				unitNumber         => $_->{'unitNumber'}
			);

			push @device_config_spec, newspec($dev);
		}
	}

	if (defined $device_ref->{'VirtualLsiLogicSASController'}) {
		my $devices = $device_ref->{'VirtualLsiLogicSASController'};

		foreach (@$devices) {
			# Create VirtualLsiLogicSASController
			my $dev = VirtualLsiLogicSASController->new(
				busNumber          => $_->{'busNumber'},
				controllerKey      => $_->{'controllerKey'},
				hotAddRemove       => $_->{'hotAddRemove'},
				key                => $_->{'key'},
				scsiCtlrUnitNumber => $_->{'scsiCtlrUnitNumber'},
				sharedBus          => VirtualSCSISharing->new('noSharing'),
				unitNumber         => $_->{'unitNumber'}
			);

			push @device_config_spec, newspec($dev);
		}
	}

	if (defined $device_ref->{'VirtualAHCIController'}) {
		# Supported from vSphere 5.5 and Unitrends Version 7.4
		my $devices = $device_ref->{'VirtualAHCIController'};

		foreach (@$devices) {
			# Create VirtualAHCIController
			my $dev = VirtualAHCIController->new(
				busNumber          => $_->{'busNumber'},
				controllerKey      => $_->{'controllerKey'},
				key                => $_->{'key'},
				unitNumber         => $_->{'unitNumber'}
			);

			push @device_config_spec, newspec($dev);
		}
	}

	#############
	# hard disks
	#############

	if (defined $device_ref->{'VirtualDisk'}) {
		my $devices = $device_ref->{'VirtualDisk'};

		foreach (@$devices) {
			# Bug 16541 - Restore metadata or not for the excluded disk
			if ((defined $_->{'is_excluded'}) && ($metadata eq "NO")) {
				print STDERR "User doesn't want to restore metadata for this VM, since metadata value passed was ". $metadata . "\n";
				print STDERR "Skipping disk with key ".$_->{'key'}."\n";
				next;
			}			

			# Create VirtualDisk devices
			my $backing;
			if (not defined $_->{'backing'}{'_backing_type'}) {
				confess "_backing_type not defined." . Dumper($_);
			}

			my $disk_filename = $datastore_path;

			if ($instant_mode) {
				# Look up filename for key.
				print STDERR "Disk map: " . Dumper($disk_map_href) . "\n";
				my $key = $_->{'key'};
				my $filename = $datastore_path . ' ' . $restore_dir . '/' . $disk_map_href->{$key};
				print STDERR "Filename for $key is $filename.\n";

				$disk_filename = $filename;
				print STDERR "Disk filename is $disk_filename (same as above)\n";
			}


			my $new_moref = bless (
					{ value => $datastore_moref->{'value'},
					  type => $datastore_moref->{'type'} } ,
					'ManagedObjectReference');
			if ($_->{'backing'}{'_backing_type'} eq 'VirtualDiskFlatVer2BackingInfo') {
				$backing = VirtualDiskFlatVer2BackingInfo->new(
					datastore       => $new_moref,
					diskMode        => $_->{'backing'}{'diskMode'},
					fileName        => $disk_filename,
					thinProvisioned => $_->{'backing'}{'thinProvisioned'},
					# skip UUID
					writeThrough    => $_->{'backing'}{'writeThrough'}
				);
			} elsif ($_->{'backing'}{'_backing_type'} eq 'VirtualDiskRawDiskMappingVer1BackingInfo') {
				# restore as a standard virtual disk
				$backing = VirtualDiskFlatVer2BackingInfo->new(
					datastore	=> $new_moref,
					diskMode	=> 'persistent',
					fileName	=> $disk_filename
				);
			} else {
				carp Dumper($_);
				die "Can't identify disk backing type: " . $_->{'backing'}{'_backing_type'} . "\n";
			}
			my $dev = VirtualDisk->new(
				capacityInKB  => $_->{'capacityInKB'},
				controllerKey => $_->{'controllerKey'},
				key           => $_->{'key'},
				unitNumber    => $_->{'unitNumber'},
				backing       => $backing,
			);

			if ($instant_mode) {
				push @device_config_spec, newspec($dev);
			} else {
				push @device_config_spec, newdiskspec($dev);
			}
		}
	}

	##################
	# removable disks
	##################

	if (defined $device_ref->{'VirtualCdrom'}) {
		my $devices = $device_ref->{'VirtualCdrom'};

		foreach (@$devices) {
			my $backing;
			if ($_->{'backing'}{'_backing_type'} eq 'VirtualCdromIsoBackingInfo') {
				if (!$wir_mode) {
				# Restore as Remote backing since the .iso might not exist anymore.
				$backing = VirtualCdromRemotePassthroughBackingInfo->new(
					deviceName => "",
					exclusive => 0
				);
				} else {
					# When creating a VM for WIR, the VM needs to boot from ISO
					$backing = VirtualCdromIsoBackingInfo->new(
						datastore => $nas_ds_moref,
						fileName => $_->{'backing'}{'fileName'}
					);
				}
			} elsif ($_->{'backing'}{'_backing_type'} eq 'VirtualCdromAtapiBackingInfo') {
				$backing = VirtualCdromAtapiBackingInfo->new(
					deviceName => $_->{'backing'}{'deviceName'},
					useAutoDetect => $_->{'backing'}{'useAutoDetect'}
				);
			} elsif ($_->{'backing'}{'_backing_type'} eq 'VirtualCdromRemoteAtapiBackingInfo') {
				$backing = VirtualCdromRemoteAtapiBackingInfo->new(
					deviceName    => $_->{'backing'}{'deviceName'},
					useAutoDetect => $_->{'backing'}{'useAutoDetect'}
				);
			} elsif ($_->{'backing'}{'_backing_type'} eq 'VirtualCdromRemotePassthroughBackingInfo') {
				$backing = VirtualCdromRemotePassthroughBackingInfo->new(
					deviceName    => $_->{'backing'}{'deviceName'},
					exclusive     => $_->{'backing'}{'exclusive'},
					useAutoDetect => $_->{'backing'}{'useAutoDetect'}
				);
			} else {
				die "Can't identify Cdrom backing type: " . $_->{'backing'}{'_backing_type'} . "\n";
			}

			my $dev = VirtualCdrom->new(
				controllerKey => $_->{'controllerKey'},
				key           => $_->{'key'},
				unitNumber    => $_->{'unitNumber'},
				backing       => $backing,
				connectable   => VirtualDeviceConnectInfo->new(
						allowGuestControl => $_->{'connectable'}{'allowGuestControl'},
						connected         => $_->{'connectable'}{'connected'},
						# connected => '1'
						startConnected    => $_->{'connectable'}{'startConnected'}
				)
			);

			push @device_config_spec, newspec($dev);
		}
	}

	if (defined $device_ref->{'VirtualFloppy'}) {
		my $devices = $device_ref->{'VirtualFloppy'};

		foreach (@$devices) {
			# Create VirtualFloppy devices
			my $backing;
			if (not defined $_->{'backing'}{'_backing_type'}) {
				confess "_backing_type not defined." . Dumper($_);
			}
			if ($_->{'backing'}{'_backing_type'} eq 'VirtualFloppyRemoteDeviceBackingInfo') {
				$backing = VirtualFloppyRemoteDeviceBackingInfo->new(
					deviceName    => $_->{'backing'}{'deviceName'},
					useAutoDetect => $_->{'backing'}{'useAutoDetect'}
				);
			} elsif ($_->{'backing'}{'_backing_type'} eq 'VirtualFloppyImageBackingInfo') {
				if (!$wir_mode) {
					print STDERR Dumper ($datastore_moref);
					# This was a file-backed floppy -- restore it as a client device
					$backing = VirtualFloppyRemoteDeviceBackingInfo->new(
						deviceName    => "floppy",
						useAutoDetect => 0
					);
					# add information to annotation
					$$annotation_ref = "Warning: This VM had a Virtual Floppy device backed by " .
						$_->{'backing'}->{'fileName'} .
						" which has been restored as a client device.  Reconfiguration may be necessary.\n" .
						$$annotation_ref;
				} else {
					# When creating a VM for WIR, the VM needs the Virtual Floppy as is
					$backing = VirtualFloppyImageBackingInfo->new(
						datastore => $nas_ds_moref,
						fileName => $_->{'backing'}{'fileName'}
					);
				}
			} elsif ($_->{'backing'}{'_backing_type'} eq 'VirtualFloppyDeviceBackingInfo') {
				# This was a physical floppy on the host -- restore it as a client device
				$backing = VirtualFloppyRemoteDeviceBackingInfo->new(
					deviceName    => $_->{'backing'}{'deviceName'},
					useAutoDetect => $_->{'backing'}{'useAutoDetect'}
				);
				# add information to annotation
				$$annotation_ref = "Warning: This VM had a Virtual Floppy device backed by " .
					$_->{'backing'}->{'deviceName'} .
					" which has been restored as a client device.  Reconfiguration may be necessary.\n" .
					$$annotation_ref;
			} else {
				carp Dumper($_);
				die "Can't identify floppy backing type: " . $_->{'backing'}{'_backing_type'} . "\n";
			}

			my $dev = VirtualFloppy->new(
				controllerKey => $_->{'controllerKey'},
				key           => $_->{'key'},
				unitNumber    => $_->{'unitNumber'},
				backing       => $backing,
				connectable   => VirtualDeviceConnectInfo->new(
						allowGuestControl => $_->{'connectable'}{'allowGuestControl'},
						connected         => $_->{'connectable'}{'connected'},
						startConnected    => $_->{'connectable'}{'startConnected'}
				)
			);

			push @device_config_spec, newspec($dev);
		}
	}

	##########
	# network
	##########

	if (defined $device_ref->{'VirtualE1000'}) {
		my $devices = $device_ref->{'VirtualE1000'};

		foreach (@$devices) {
			# Create VirtualE1000 device

			if (!(defined $_->{'backing'}{'_backing_type'}) and !(defined $_->{'backing'}{'useAutoDetect'})) {
				# In 5.0 and 5.0.1 we didn't back up enough metadata to restore this device.
				# For those backups, do not restore the E1000 NIC.
				# (5.0.1 and earlier would not have saved _backing_type, either.)
				print STDERR "Skipping restore of VirtualE1000 NIC.\n";
				$$annotation_ref = "Warning: This VM was previously configured with a Virtual E1000 Ethernet adapter which has not been restored.  Reconfiguration may be necessary.\n" .
					$$annotation_ref;

				next;
			}

			my $backing = assemble_nic_backing(\$_);
			if (not defined $backing) {
				print STDERR "Skipping VirtualE1000 restore because backing is not defined.\n";
				$$annotation_ref = "Warning: This VM was previously configured with a Virtual E1000 Ethernet adapter which has not been restored.  Reconfiguration may be necessary.\n" .
					$$annotation_ref;
				next;
			}

			my $dev = VirtualE1000->new(
				addressType      => $_->{'addressType'},
				controllerKey    => $_->{'controllerKey'},
				key              => $_->{'key'},
				macAddress       => $_->{'macAddress'},
				unitNumber       => $_->{'unitNumber'},
				wakeOnLanEnabled => $_->{'wakeOnLanEnabled'},
				backing          => $backing,
				connectable => VirtualDeviceConnectInfo->new(
					allowGuestControl => $_->{'connectable'}{'allowGuestControl'},
					connected         => $_->{'connectable'}{'connected'},
					startConnected    => $_->{'connectable'}{'startConnected'}
				)
			);

			push @device_config_spec, newspec($dev);
		}
	}

	if (defined $device_ref->{'VirtualE1000e'}) {
		my $devices = $device_ref->{'VirtualE1000e'};

		foreach (@$devices) {
			# Create Virtual E1000e device

			my $backing = assemble_nic_backing(\$_);
			if (not defined $backing) {
				print STDERR "Skipping VirtualE1000e restore because backing is not defined.\n";
				$$annotation_ref = "Warning: This VM was previous configured with a Virtual E1000e Ethernet adapter which has not been restored.  Reconfiguration may be necessary.\n" . $$annotation_ref;
				next;
			}
			
			my $dev = VirtualE1000e->new(
				addressType         => $_->{'addressType'},
				controllerKey       => $_->{'controllerKey'},
				key                 => $_->{'key'},
				macAddress          => $_->{'macAddress'},
				unitNumber          => $_->{'unitNumber'},
				wakeOnLanEnabled    => $_->{'wakeOnLanEnabled'},
				backing             => $backing,
				connectable => VirtualDeviceConnectInfo->new(
					allowGuestControl => $_->{'connectable'}{'allowGuestControl'},
					connected         => $_->{'connectable'}{'connected'},
					startConnected    => $_->{'connectable'}{'startConnected'}
				)
			);

			push @device_config_spec, newspec($dev);
		}
	}

	if (defined $device_ref->{'VirtualPCNet32'}) {
		my $devices = $device_ref->{'VirtualPCNet32'};	

		foreach (@$devices) {
			# Create VirtualPCNet32 device
			my $backing = assemble_nic_backing(\$_);
			if (not defined $backing) {
				print STDERR "Skipping VirtualPCNet32 restore because backing is not defined.\n";
				$$annotation_ref = "Warning: This VM was previously configured with a Virtual PCNet32 Ethernet adapter which has not been restored.  Reconfiguration may be necessary.\n" .
					$$annotation_ref;
				next;
			}

			my $dev = VirtualPCNet32->new(
				addressType      => $_->{'addressType'},
				controllerKey    => $_->{'controllerKey'},
				key              => $_->{'key'},
				macAddress       => $_->{'macAddress'},
				unitNumber       => $_->{'unitNumber'},
				wakeOnLanEnabled => $_->{'wakeOnLanEnabled'},
				backing          => $backing,
				connectable => VirtualDeviceConnectInfo->new(
					allowGuestControl => $_->{'connectable'}{'allowGuestControl'},
					connected         => $_->{'connectable'}{'connected'},
					startConnected    => $_->{'connectable'}{'startConnected'},
				)
			);

			push @device_config_spec, newspec($dev);
		}

	}

	if (defined $device_ref->{'VirtualVmxnet'}) {
		my $devices = $device_ref->{'VirtualVmxnet'};

		foreach (@$devices) {
			# Create VirtualVmxnet device

			my $backing = assemble_nic_backing(\$_);
			if (not defined $backing) {
				print STDERR "Skipping VirtualVmxnet restore because backing is not defined.\n";
				$$annotation_ref = "Warning: This VM was previously configured with a Virtual Vmxnet Ethernet adapter which has not been restored.  Reconfiguration may be necessary.\n" .
					$$annotation_ref;
				next;
			}

			my $dev = VirtualVmxnet->new(
				addressType      => $_->{'addressType'},
				controllerKey    => $_->{'controllerKey'},
				key              => $_->{'key'},
				macAddress       => $_->{'macAddress'},
				unitNumber       => $_->{'unitNumber'},
				wakeOnLanEnabled => $_->{'wakeOnLanEnabled'},
				backing          => $backing,
				connectable => VirtualDeviceConnectInfo->new(
					allowGuestControl => $_->{'connectable'}{'allowGuestControl'},
					connected         => $_->{'connectable'}{'connected'},
					startConnected    => $_->{'connectable'}{'startConnected'},
				)
			);

			push @device_config_spec, newspec($dev);
		}
	}

	if (defined $device_ref->{'VirtualVmxnet2'}) {
		my $devices = $device_ref->{'VirtualVmxnet2'};

		foreach (@$devices) {
			# Create VirtualVmxnet2 device

			my $backing = assemble_nic_backing(\$_);
			if (not defined $backing) {
				print STDERR "Skipping VirtualVmxnet2 restore because backing is not defined.\n";
				$$annotation_ref = "Warning: This VM was previously configured with a Virtual Vmxnet2 Ethernet adapter which has not been restored.  Reconfiguration may be necessary.\n" .
					$$annotation_ref;
				next;
			}

			my $dev = VirtualVmxnet2->new(
				addressType      => $_->{'addressType'},
				controllerKey    => $_->{'controllerKey'},
				key              => $_->{'key'},
				macAddress       => $_->{'macAddress'},
				unitNumber       => $_->{'unitNumber'},
				wakeOnLanEnabled => $_->{'wakeOnLanEnabled'},
				backing          => $backing,
				connectable => VirtualDeviceConnectInfo->new(
					allowGuestControl => $_->{'connectable'}{'allowGuestControl'},
					connected         => $_->{'connectable'}{'connected'},
					startConnected    => $_->{'connectable'}{'startConnected'},
				)
			);

			push @device_config_spec, newspec($dev);
		}
	}

	if (defined $device_ref->{'VirtualVmxnet3'}) {
		my $devices = $device_ref->{'VirtualVmxnet3'};

		foreach (@$devices) {
			# Create VirtualVmxnet3 device

			my $backing = assemble_nic_backing(\$_);
			if (not defined $backing) {
				print STDERR "Skipping Virtual Vmxnet3 restore because backing is not defined.\n";
				$$annotation_ref = "Warning: This VM was previously configured with a Virtual Vmxnet3 Ethernet adapter which has not been restored.  Reconfiguration may be necessary.\n" .
					$$annotation_ref;
				next;
			}

			my $dev = VirtualVmxnet3->new(
				addressType      => $_->{'addressType'},
				controllerKey    => $_->{'controllerKey'},
				key              => $_->{'key'},
				macAddress       => $_->{'macAddress'},
				unitNumber       => $_->{'unitNumber'},
				wakeOnLanEnabled => $_->{'wakeOnLanEnabled'},
				backing          => $backing,
				connectable => VirtualDeviceConnectInfo->new(
					allowGuestControl => $_->{'connectable'}{'allowGuestControl'},
					connected         => $_->{'connectable'}{'connected'},
					startConnected    => $_->{'connectable'}{'startConnected'},
				)
			);

			push @device_config_spec, newspec($dev);
		}
	}

	if (defined $device_ref->{'VirtualSriovEthernetCard'}) {
		my $devices = $device_ref->{'VirtualSriovEthernetCard'};

		foreach (@$devices) {
			# Create VirtualSriovEthernetCard device

			my $backing = assemble_nic_backing(\$_);
			if (not defined $backing) {
				print STDERR "Skipping Virtual Sriov Ethernet Card restore because backing is not defined.\n";
				$$annotation_ref = "Warning: This VM was previously configured with a Virtual Sriov  Ethernet adapter which has not been restored.  Reconfiguration may be necessary.\n" .
					$$annotation_ref;
				next;
			}

			my $dev = VirtualSriovEthernetCard->new(
				addressType      => $_->{'addressType'},
				controllerKey    => $_->{'controllerKey'},
				key              => $_->{'key'},
				macAddress       => $_->{'macAddress'},
				unitNumber       => $_->{'unitNumber'},
				wakeOnLanEnabled => $_->{'wakeOnLanEnabled'},
				backing          => $backing,
				connectable => VirtualDeviceConnectInfo->new(
					allowGuestControl => $_->{'connectable'}{'allowGuestControl'},
					connected         => $_->{'connectable'}{'connected'},
					startConnected    => $_->{'connectable'}{'startConnected'},
				)
			);

			push @device_config_spec, newspec($dev);
		}
	}

	#######
	# misc
	#######

	if (defined $device_ref->{'VirtualUSBController'}) {
		my $devices = $device_ref->{'VirtualUSBController'};

		foreach (@$devices) {
			my $dev = VirtualUSBController->new(
				autoConnectDevices => $_->{'autoConnectDevices'},
				busNumber          => $_->{'busNumber'},
				ehciEnabled        => $_->{'ehciEnabled'},
				key                => $_->{'key'},
				unitNumber         => $_->{'unitNumber'}
			);
			
			push @device_config_spec, newspec($dev);
		}
	}

	if (defined $device_ref->{'VirtualUSBXHCIController'}) {
		my $devices = $device_ref->{'VirtualUSBXHCIController'};

		foreach (@$devices) {
			my $dev = VirtualUSBXHCIController->new(
				autoConnectDevices => $_->{'autoConnectDevices'},
				busNumber          => $_->{'busNumber'},
				key                => $_->{'key'},
				unitNumber         => $_->{'unitNumber'}
			);

			push @device_config_spec, newspec($dev);
		}
	}

	if (defined $device_ref->{'VirtualParallelPort'}) {
		my $devices = $device_ref->{'VirtualParallelPort'};

		# VirtualParallelPorts are backed up but not restored.
		# Use the information from the backup to generate a warning message for the VM annotation.
		
		foreach (@$devices) {
			if ($_->{'backing'}{'_backing_type'} eq 'VirtualParallelPortDeviceBackingInfo') {
				$$annotation_ref = "Warning: This VM had a Virtual Parallel Port device backed by device \"" .
					$_->{'backing'}{'deviceName'} .
					"\" which has not been restored.  Reconfiguration may be necessary.\n" .
					$$annotation_ref;
				next;
			} elsif ($_->{'backing'}{'_backing_type'} eq 'VirtualParallelPortFileBackingInfo') {
				$$annotation_ref = "Warning: This VM had a Virtual Parallel Port device backed by file \"" .
					$_->{'backing'}{'fileName'} .
					"\" which has not been restored.  Reconfiguration may be necessary.\n" .
					$$annotation_ref;
				next;
			} else {
				print STDERR Dumper($_);
				die "Cannot determine backing type for Parallel port: " . $_->{'backing'}{'_backing_type'} . "\n";
			}
		}	
	}

	if (defined $device_ref->{'VirtualSCSIPassthrough'}) {
		my $devices = $device_ref->{'VirtualSCSIPassthrough'};

		foreach (@$devices) {
			my $backing = "unknown!";

			if ($_->{'backing'}{'_backing_type'} eq 'VirtualSCSIPassthroughDeviceBackingInfo') {
				$backing = VirtualSCSIPassthroughDeviceBackingInfo->new(
						deviceName    => $_->{'backing'}{'deviceName'},
						useAutoDetect => $_->{'backing'}{'useAutoDetect'}
				);
			}

			my $connectable = VirtualDeviceConnectInfo->new(
					allowGuestControl => $_->{'connectable'}{'allowGuestControl'},
					connected         => $_->{'connectable'}{'connected'},
					startConnected    => $_->{'connectable'}{'startConnected'}
			);

			my $dev = VirtualSCSIPassthrough->new(
				controllerKey => $_->{'controllerKey'},
				key           => $_->{'key'},
				unitNumber    => $_->{'unitNumber'},
				backing       => $backing,
				connectable   => $connectable
			);

			push @device_config_spec, newspec($dev);
		}
	}

	if (defined $device_ref->{'VirtualSerialPort'}) {
		my $devices = $device_ref->{'VirtualSerialPort'};

		# VirtualSerialPorts are backed up but not restored.
		# Use metadata from backup to generate a message for the VM annotation.

		foreach (@$devices) {
			if ($_->{'backing'}{'_backing_type'} eq 'VirtualSerialPortDeviceBackingInfo') {
				$$annotation_ref = "Warning: This VM had a Virtual Serial Port device backed by device \"" .
					$_->{'backing'}{'deviceName'} .
					"\" which has not been restored.  Reconfiguration may be necessary.\n" .
					$$annotation_ref;
				next;
			} elsif ($_->{'backing'}{'_backing_type'} eq 'VirtualSerialPortFileBackingInfo') {
				$$annotation_ref = "Warning: This VM had a Virtual Serial Port device backed by file \"" .
					$_->{'backing'}{'fileName'} .
					"\" which has not been restored.  Reconfiguration may be necessary.\n" .
					$$annotation_ref;
				next;
			} elsif ($_->{'backing'}{'_backing_type'} eq 'VirtualSerialPortPipeBackingInfo') {
				$$annotation_ref = "Warning: This VM had a Virtual Serial Port device backed by pipe \"" .
					$_->{'backing'}{'pipeName'} .
					"\" which has not been restored.  Reconfiguration may be necessary.\n" .
					$$annotation_ref;
				next;
			} elsif ($_->{'backing'}{'_backing_type'} eq 'VirtualSerialPortURIBackingInfo') {
				$$annotation_ref = "Warning: This VM had a Virtual Serial Port device backed by URI \"" .
					$_->{'backing'}{'serviceURI'} .
					"\" which has not been restored.  Reconfiguration may be necessary.\n" .
					$$annotation_ref;
				next;
			} else {
				die "Cannot determine type of backing for VirtualSerialPort: " . Dumper($_) . "\n";
			}
		}
	}

	if (defined $device_ref->{'VirtualMachineVMIROM'}) {
		my $devices = $device_ref->{'VirtualMachineVMIROM'};
		#print STDERR "Found VMIROM\n";
		foreach (@$devices) {
			#print STDERR "Editing annotation for VMIROM\n";
			$$annotation_ref = "Warning: This VM was previously configured for VMI virtualization.  This configuration has not been restored. Reconfiguration may be necessary.\n" . $$annotation_ref;
			last;
		}
	}

	if (defined $device_ref->{'VirtualHdAudioCard'}) {
		my $devices = $device_ref->{'VirtualHdAudioCard'};
		foreach (@$devices) {
			my $backing = "unknown!";

			if ($_->{'backing'}{'_backing_type'} eq 'VirtualSoundCardDeviceBackingInfo') {
				$backing = VirtualSoundCardDeviceBackingInfo->new(
					deviceName    => $_->{'backing'}{'deviceName'},
					useAutoDetect => $_->{'backing'}{'useAutoDetect'}
				);
			} else {
				die 'Could not identify VirtualHdAudioCard backing type' . $_->{'backing'}{'_backing_type'} . "\n";
			}

			my $dev = VirtualHdAudioCard->new(
				backing => $backing,
				key     => $_->{'key'}
			);

			push @device_config_spec, newspec($dev);
		}
	}

	if (defined $device_ref->{'VirtualEnsoniq1371'}) {
		my $devices = $device_ref->{'VirtualEnsoniq1371'};
		foreach (@$devices) {
			my $backing = "unknown!";

			if ($_->{'backing'}{'_backing_type'} eq 'VirtualSoundCardDeviceBackingInfo') {
				$backing = VirtualSoundCardDeviceBackingInfo->new(
					deviceName    => $_->{'backing'}{'deviceName'},
					useAutoDetect => $_->{'backing'}{'useAutoDetect'}
				);
			} else {
				die 'Could not identify VirtualEnsoniq1371 backing type' . $_->{'backing'}{'_backing_type'} . "\n";
			}

			my $dev = VirtualEnsoniq1371->new(
				backing => $backing,
				key     => $_->{'key'}
			);

			push @device_config_spec, newspec($dev);
		}
	}

	if (defined $device_ref->{'VirtualSoundBlaster16'}) {
		my $devices = $device_ref->{'VirtualSoundBlaster16'};
		foreach (@$devices) {
			my $backing = "unknown!";

			if ($_->{'backing'}{'_backing_type'} eq 'VirtualSoundCardDeviceBackingInfo') {
				$backing = VirtualSoundCardDeviceBackingInfo->new(
					deviceName    => $_->{'backing'}{'deviceName'},
					useAutoDetect => $_->{'backing'}{'useAutoDetect'}
				);
			} else {
				die 'Could not identify VirtualSoundBlaster16 backing type' . $_->{'backing'}{'_backing_type'} . "\n";
			}

			my $dev = VirtualSoundBlaster16->new(
				backing => $backing,
				key     => $_->{'key'}
			);

			push @device_config_spec, newspec($dev);
		}
	}

	#if (defined $device_ref->{'VirtualMachineVideoCard'}) {
	#	my $devices = $device_ref->{'VirtualMachineVideoCard'};
	#	foreach (@$devices) {
	#		my $opts = {};
	#		$opts->{'controllerKey'} = $_->{'controllerKey'};
	#		$opts->{'key'}           = $_->{'key'};
	#		$opts->{'unitNumber'}    = $_->{'unitNumber'};
	#
	#		# Optional keys
	#		foreach my $key ('enable3DSupport', 'numDisplays', 'useAutoDetect', 'videoRamSizeInKB') {
	#			if (defined($_->{$key})) {
	#				if ($_->{$key} ne "") {
	#					$opts->{$key} = $_->{$key};
	#				}
	#			}
	#		}
	#
	#		my $dev = VirtualMachineVideoCard->new(%$opts);
	#
	#		push @device_config_spec, newspec($dev);
	#	}
	#}


	# The following device types should not be restored:
	#  VirtualMachineVMCIDevice
	#  VirtualPS2Controller
	#  VirtualSIOController
	
	return \@device_config_spec;
}

sub find_datacenter {
	# Given an object view, find the datacenter in the parent chain
	# Return a view for this datacenter

	my $start_view = shift;

	my $cur_view = $start_view;
	while (defined $cur_view) {
		my $next_view = Vim::get_view(mo_ref => $cur_view->parent);

		if ($next_view->isa('Datacenter')) {
			return $next_view;
		}

		$cur_view = $next_view;
	}

	die "Datacenter not found for object " . Dumper($start_view) . "\n";
}


sub create_vm {
	# Create a new VM, return its moref.

	my $vmxml = shift;
	my $datastore_name = shift;
	my $hostname = shift;
	my $new_vm_name = shift;
	my $metadata = shift;
	my $pool = shift;
	my $wir_mode = shift;

	my $ir_opts_href = shift;  # Defined for Instant Recovery only.
	my $instant_rec_mode = 0;
	my $ir_datastore_moref_ref;
	my $ir_datastore_moref;
	my $ir_datastore_name;
	my $disk_map_href;
	my $restore_dir;
	my $ir_audit = 0;
	if (defined ($ir_opts_href)) {
		$instant_rec_mode = 1;
		$ir_datastore_moref_ref = $ir_opts_href->{'ir_datastore_moref'};
		$ir_datastore_moref = $$ir_datastore_moref_ref;
		$ir_datastore_name = $ir_opts_href->{'datastore_name'};
		$disk_map_href = $ir_opts_href->{'disk_map_href'};
		$restore_dir = $ir_opts_href->{'restore_dir'};
		$ir_audit = $ir_opts_href->{'ir_audit'};

		print STDERR "Restore dir is $restore_dir\n";
	}

	my $bootOptionsSpec;
	my $new_vm_moref;
	
	{
		# get all of the metadata we need to recreate the VM
		my @vms = $vmxml->getElementsByTagName('virtual_machine');
		my $vm = $vms[0];
		my $name = $new_vm_name;
		my $uuid = $vm->findvalue('uuid');
		my $version = $vm->findvalue('version');
		my $guestId = $vm->findvalue('guestId');
		my $annotation = $vm->findvalue('annotation');
		my $firmware = $vm->findvalue('firmware');

		my $cpuHotAddEnabled             = $vm->findvalue('cpuHotAddEnabled');
		my $cpuHotRemoveEnabled          = $vm->findvalue('cpuHotRemoveEnabled');
		my $maxMksConnections            = $vm->findvalue('maxMksConnections');
		my $memoryHotAddEnabled          = $vm->findvalue('memoryHotAddEnabled');
		my $memoryReservationLockedToMax = $vm->findvalue('memoryReservationLockedToMax');

		my $original_name = $vm->findvalue('name');

		# Check to see if we have any 3+ byte characters in any of these strings, 
		# if so, replace them with an underscore '_'.
		$new_vm_name =~ s/[^\x00-\x7f]/_/g;
		$annotation =~ s/[^\x00-\x7f]/_/g;
		$original_name =~ s/[^\x00-\x7f]/_/g;
		# Bug 16153 - Inserting this to handle non-ASCII characters in the restore VM name
		$name =~ s/[^\x00-\x7f]/_/g;

		# Configuration
		my @extraConfig;
		foreach my $exconfig ($vm->findnodes('config/extraConfig')) {
			foreach my $conf ($exconfig->childNodes) {
				if ($conf->nodeName eq 'cpuid.coresPerSocket') {
					my $ov = OptionValue->new(
							key => 'cpuid.coresPerSocket',
							value => $conf->textContent
					);

					unshift @extraConfig, $ov;
				}
			}
			if ($instant_rec_mode and $ir_audit) {
				# This will help to identify the VM that gets created for VMware IR in Audit Mode
				# ********* NOTE - This field is used by RDR for the VM identification **********
				my $unitrends_id = OptionValue->new(
						key => 'unitrends.vm.type',
						value => 'instant_recovery_audit'
						);
				unshift @extraConfig, $unitrends_id;
			}
		}

		# Basic hardware information
		my @hardwares = $vm->getElementsByTagName('hardware');
		my $hardware = $hardwares[0];
		my $memoryMB = $hardware->findvalue('memoryMB');
		my $numCPUs = $hardware->findvalue('numCPU');

		# Find the datastore's moref
		my %datastore_info = get_datastore_info($datastore_name, $hostname);
		my $datastore_moref = $datastore_info{'mor'};
		if ($datastore_moref eq 0) {
			die "Error accessing datastore information.\n";
		}

		# Create the datastore's path
		my $datastore_path = "[" . $datastore_info{'name'} . "]";
		print STDERR "Datastore path: " . $datastore_path . "\n";
		
		my $ir_datastore_path;
		if ($instant_rec_mode) {
			$ir_datastore_path = '[' . $ir_datastore_name . ']';
			print STDERR "IR Datastore path: " . $ir_datastore_path . "\n";
		}

		# Get device information
		my @devices = $hardware->getChildrenByTagName('devices');
		my @device_change;
		my $device_change_ref = \@device_change;
		foreach (@devices) {
			foreach ($_->childNodes) {
				if ($_->nodeType == XML_TEXT_NODE) {
					next;
				}

				my @device = process_device($_);
			}
		}


		my $change_spec;
		if ($instant_rec_mode) {
			$change_spec = create_device_change_spec($devices_hash, $ir_datastore_moref, $ir_datastore_path, $metadata, $wir_mode, \$annotation, $disk_map_href, $restore_dir);
		} else {  # Normal restore
			$change_spec = create_device_change_spec($devices_hash, $datastore_moref, $datastore_path, $metadata, $wir_mode, \$annotation);
		}

		print STDERR "Change spec is -\n" . Dumper($change_spec) . "\n";

		# Specify where the VM is located, let log/snap/suspend dirs default
		my $files = VirtualMachineFileInfo->new(
			logDirectory      => undef,
			snapshotDirectory => undef,
			suspendDirectory  => undef,
			vmPathName => $datastore_path
		);

		# Create the configuration spec for the entire VM.
		my $opts = ();
		$opts->{'name'}        = $name;
		$opts->{'memoryMB'}    = $memoryMB;
		$opts->{'files'}       = $files;
		$opts->{'numCPUs'}     = $numCPUs;
		$opts->{'guestId'}     = $guestId;
		$opts->{'deviceChange'} = $change_spec;
		$opts->{'extraConfig'} = \@extraConfig;
		if (defined $firmware) {
			if ($firmware ne "") {
				$opts->{'firmware'}    = $firmware;
			}
		}
		if (defined $cpuHotAddEnabled) {
			if ($cpuHotAddEnabled ne "") {
				$opts->{'cpuHotAddEnabled'} = $cpuHotAddEnabled;
			}
		}
		if (defined $cpuHotRemoveEnabled) {
			if ($cpuHotRemoveEnabled ne "") {
				$opts->{'cpuHotRemoveEnabled'} = $cpuHotRemoveEnabled;
			}
		}
		if (defined $maxMksConnections) {
			if ($maxMksConnections ne "") {
				$opts->{'maxMksConnections'} = $maxMksConnections;
			}
		}
		if (defined $memoryHotAddEnabled) {
			if ($memoryHotAddEnabled ne "") {
				$opts->{'memoryHotAddEnabled'} = $memoryHotAddEnabled;
			}
		}
		if (defined $memoryReservationLockedToMax) {
			if ($memoryReservationLockedToMax ne "") {
				if ($memoryReservationLockedToMax eq 1) {
					# Bug 19996 - Always restore without memory reservation lock
					$memoryReservationLockedToMax = 0;
					$annotation = "Warning: This VM was previously configured with memory " .
					"reservation. This configuration has not been restored. Reconfiguration " .
					"may be necessary.\n" . $annotation;
				}
				$opts->{'memoryReservationLockedToMax'} = $memoryReservationLockedToMax;
			}
		}

		my $vm_config_spec = create_vm_config_spec($opts);

		my $indent_old = $Data::Dumper::Indent;
		$Data::Dumper::Indent = 3;
		print STDERR "vm_config_spec: " . Dumper($vm_config_spec);
		$Data::Dumper::Indent = $indent_old;

		# Add a timestamp to the Annotations.
		my $thetime = strftime("%H:%M:%S %Z %a %b %e %Y", localtime);
		$vm_config_spec->{'annotation'} = "Restored from a Unitrends DPU at " . $thetime . ".\n" .
		   "Original VM name was: " . $original_name . ".\n\n" . $annotation;

		# Find the host we're restoring to
		my $host_view = Vim::find_entity_view(
			view_type => 'HostSystem',
			filter => { 'name' => $hostname }
		);

		# Find the datacenter and compute resource we're restoring to
		my $datacenter = find_datacenter($host_view);
		my $vm_folder_view = Vim::get_view(mo_ref => $datacenter->vmFolder);
		my $comp_res_view = Vim::get_view(mo_ref => $host_view->parent);

		my $found_pool = "";
		my $found_pool_view;
		my $rp_view;
		my $resource_pools;
		my @recur_rp;
		my $rp_key;
		my $flag = 0;

		# If empty argument was passed, directly restore to the ESX Server
		if ($pool ne 'empty') {
			print STDERR "Value of pool is $pool\n";
			# Find a view for traversing all the RPs on the ESX
			$rp_view = Vim::find_entity_view(
					view_type => 'ResourcePool',
					begin_entity => $comp_res_view->resourcePool
					);
			$resource_pools = $rp_view->resourcePool;

			foreach (@$resource_pools) {
				$rp_view = Vim::get_view( mo_ref => $_, properties => [ 'resourcePool' ] );
				$rp_key = $rp_view->get_property('mo_ref')->value;

				push (@recur_rp, $_);
				if (defined ($rp_view->resourcePool)) {
					@recur_rp = rp_view_restore ($rp_view->resourcePool, \@recur_rp);
				}
			}

			foreach (@recur_rp) {
				$rp_view = Vim::get_view( mo_ref => $_, properties => [ 'resourcePool' ] );
				$rp_key = $rp_view->get_property('mo_ref')->value;
				# Exit if required 'key' of the RP is found
				if ($rp_key eq $pool) {
					$found_pool = $_;
					last;
				}
			}
		}

		if ($found_pool eq "") {
			# If the argument passed was 'empty' or invalid RP key was passed, restore the VM to the ESX Server itself
			# Log the finding, but don't fail
			print STDERR "Pool argument passed was $pool\n";
			print STDERR "Invalid or 'empty' pool argument for restore\n";
			print STDERR "Restoring the Virtual Machine to the ESX Server\n";
			$found_pool = $comp_res_view->resourcePool;
			$flag = 0;
		} else {
			$found_pool_view = Vim::get_view( mo_ref => $found_pool );

			# Restoring directly to a VA
			if (ref($found_pool_view) eq 'VirtualApp') {
				print STDERR "Restoring the Virtual Machine directly to the VirtualApp - ".$found_pool_view->name."\n";
				$flag = 1;
			} elsif (ref($found_pool_view) eq 'ResourcePool') {

				# Restoring to a RP, but need to find whether 
				#	- this RP is a child/grand-child of a VA or
				#	- this RP is a child/grand-child of a RP

				print STDERR "Restoring the Virtual Machine to the ResourcePool - ".$found_pool_view->name."\n";
				my $found_pool_view_parent = Vim::get_view( mo_ref => $found_pool_view->parent );

				if (ref($found_pool_view_parent) eq 'VirtualApp') {
					print STDERR "The ResourcePool ".$found_pool_view->name." is a child of the VirtualApp ".$found_pool_view_parent->name."\n";
					$flag = 1;
				}

				while (ref($found_pool_view_parent) eq 'ResourcePool') {
					$found_pool_view_parent = Vim::get_view( mo_ref => $found_pool_view_parent->parent );
					
					if (ref($found_pool_view_parent) eq 'VirtualApp') {
						print STDERR "The ResourcePool ".$found_pool_view->name." is a child/grand-child of the VirtualApp ".$found_pool_view_parent->name."\n";
						$flag = 1;
						last;
					} 
				}
	
				if ($flag eq 0) {
					print STDERR "The ResourcePool ".$found_pool_view->name." is a child/grand-child of ";
					print STDERR "another ResourcePool or the ESX Server itself\n";
				}

			} else {
				print STDERR "Couldn't identify the object for restore\n";
				print STDERR "Try restoring to ESX Server\n";
				$flag = 0;
			}
		}

		# For debugging
		print STDERR "VM config spec:\n" . Dumper($vm_config_spec) . "\n";

		# Create the VM
		eval {
			if ($flag eq 0) {
				# Create the virtual machine.
				print STDERR "Calling CreateVM\n";
				$new_vm_moref = $vm_folder_view->CreateVM(
					config => $vm_config_spec,
					host => $host_view,
					pool => $found_pool
				);
			} else {
				print STDERR "Calling CreateChildVM\n";
				$new_vm_moref = $found_pool_view->CreateChildVM(
					config => $vm_config_spec,
					host => $host_view
				);
			}
		}; if ($@) {
			# Try to throw a meaningful error message...

			my $suggestion_text =
				"Please verify that the target datastore has sufficient free space for the VM " .
				"being restored, and that the datastore is online.\n";

			if (ref($@) eq 'SoapFault') {
				my $hdr = "Error creating VM -";
				if (ref($@->detail) eq 'PlatformConfigFault') {
					die "Error creating VM - PlatformConfigFault (Invalid VM configuration): " .
							${$@->detail}{'text'} . "\n";
				} elsif (ref($@->detail) eq 'DatastoreNotWritableOnHost') {
					die "$hdr Datastore not writable on target host " . Dumper($@) . "\n";
				} elsif (ref($@->detail) eq 'InaccessibleDatastore') {
					die "$hdr Datastore is not accessible " . Dumper($@) . "\n";
				} elsif (ref($@->detail) eq 'InvalidDatastore') {
					print STDERR "$hdr Datastore was not found. These are the details -\n" . Dumper($@) . "\n";
					die "$hdr Datastore " . $@->detail->name . " was not found. Try creating the virtual machine again.\n";
				} elsif (ref($@->detail) eq 'InvalidDatastorePath') {
					print STDERR "$hdr Invalid datastore path. These are the details -\n" . Dumper($@) . "\n";
					die "$hdr Invalid Datastore path. Try creating the virtual machine again.\n";
				} elsif (ref($@->detail) eq 'InvalidDeviceSpec') {
					# Dumping $@ is ugly but I'll need the info to diagnose.
					die "$hdr InvalidDeviceSpec (Invalid device configuration): " .
							Dumper $@ . "\n";
				} elsif (ref($@->detail) eq 'DatacenterMismatch') {
					die "$hdr DatacenterMismatch (Entities do not belong to the same datacenter): " .
							${$@->detail}{'text'} . Dumper($@) . "\n";
				} elsif (ref($@->detail) eq 'HostNotConnected') {
					die "$hdr HostNotConnected (Unable to communicate with the remote host)\n";
				} elsif (ref($@->detail) eq 'InvalidState') {
					die "$hdr InvalidState (The operation is not allowed in the current state): " .
							${$@->detail}{'text'} . "\n";
				} elsif (ref($@->detail) eq 'DuplicateName') {
					die "$hdr DuplicateName (Virtual machine with name " . $@->detail->name . " already exists.)\n";
				} elsif (ref($@->detail) eq 'OutOfBounds') {
					die "$hdr OutOfBounds (MaxSupportedVMs host capability has been exceeded.)\n";
				} elsif (ref($@->detail) eq 'InvalidArgument') {
					die "$hdr InvalidArgument\n" . Dumper($@);
				} elsif (ref($@->detail) eq 'InvalidDeviceBacking') {
					die "$hdr InvalidDeviceBacking\n" . Dumper($@) . "\n" . Dumper($vm_config_spec) . "\n";
				} elsif (ref($@->detail) eq 'InvalidName') {
					die "$hdr Invalid Name: " . $@->name . "\n";
				} elsif (ref($@->detail) eq 'FileAlreadyExists') {
					die "$hdr FileAlreadyExists: " .
						$@->fault_string . "\n";
				} elsif (ref($@->detail) eq 'FileNotFound') {
					die "$hdr FileNotFound: " .
						$@->fault_string . "\n";
				} elsif (ref($@->detail) eq 'NoPermission') {
					die "$hdr NoPermission: " . "Requires permission " . $@->detail->privilegeId . "\n";
				} elsif (ref($@->detail) eq 'NotSupported') {
					die "$hdr NotSupported: " . $@->fault_string . "\n";
				} else {
					die "Error creating VM: " . $suggestion_text . $@ . "\n";
				}
			} else {
				die "Error creating VM: " . $suggestion_text . $@ . "\n";
			}
		}
	}

	return $new_vm_moref;
}

# Sub-routine to build up a complete list of RPs and VAs
sub rp_view_restore ($$);
sub rp_view_restore ($$) {
	my($recur1, $recur_rp) = @_;
	my @rp1 = @{$recur1};
	my @rp2 = @{$recur_rp};

	my $rp_view_temp;
	foreach (@rp1) {
		$rp_view_temp = Vim::get_view(
					mo_ref => $_,
					properties => [ 'name', 'resourcePool' ]
					);

		push (@rp2, $_);
		if (defined ($rp_view_temp->resourcePool)) {
			@rp2 = rp_view_restore ($rp_view_temp->resourcePool, \@rp2);
		}
	}

	return @rp2;
}

sub get_datastore_info {
	my $datastore_name = shift;
	my $host_name = shift;

	my $found_datastore = 0;
	my $mor;
	my $name;

	# The hostname has to be fully-qualified and precise for this to work.
	# We gather the hostname when creating the list of restore targets, so it can be provided here.

	my $host_view = Vim::find_entity_view(
		view_type => 'HostSystem',
		filter => { 'name' => $host_name }
	);

	if (!$host_view) {
		die "Couldn't find host_view while getting datastore info for hostname $host_name.\n";
	}

	my $datastore_morefs = $host_view->datastore;

	my $datastores = Vim::get_views(mo_ref_array => $datastore_morefs);

	if ($datastore_name) {
		foreach (@$datastores) {
			$name = $_->summary->name;
			if ($name eq $datastore_name) {
				my $ds_disksize = ($_->summary->freeSpace)/1024;

				$mor = $_->{mo_ref};
			
				$found_datastore = 1;
				last;
			}
		}
	}

	if (!$found_datastore) {
		die "Datastore " . $datastore_name . " not found\n";
		return (mor => 0, name => 'datastore_error');
	}

	return (name => $name, mor => $mor);
}

sub append_disk_filenames {
	my $restore_result = shift;
	my $new_vm_moref = shift;
	my $xml_target = shift;

	my $vm_view = Vim::get_view(mo_ref => $new_vm_moref);

	my $disk_files = VProtect::xml_get_or_create($xml_target, 'disk_files');

	my $devices = $vm_view->config->hardware->device;
	foreach (@$devices) {
		print STDERR Dumper($_);
		if ($_->isa('VirtualDisk')) {
			my $fileName = $_->backing->fileName;
			my $key = $_->key;

			my $disknode = VProtect::xml_get_or_create($xml_target, 'disk_file');
			VProtect::append_to_xml($disknode, 'key', $key);
			VProtect::append_to_xml($disknode, 'fileName', $fileName);
			
			$disk_files->addChild($disknode);
		}
	}

	$restore_result->addChild($disk_files);
}

sub check_for_maintenance_mode {
	my $hostname = shift;

	# Find the host we're restoring to
	my $host_view = Vim::find_entity_view(
		view_type => 'HostSystem',
		filter => { 'name' => $hostname },
		properties => [ 'runtime' ]
	);
	if (not defined $host_view) {
		die "Host $hostname not found, while checking maintenance mode status.\n";
	}

	if ($host_view->runtime->inMaintenanceMode) {
		die "VM cannot be restored because the ESX host is in Maintenance Mode.\n";
	}
}

sub convert_to_template {
	my $vm_uuid = shift;
	my $warning_ref = shift;

	my $vm_moref = VProtect::vim_get_vm_by_uuid($vm_uuid);
	if (!defined $vm_moref) {
		die "VM with UUID ".$vm_uuid." was not found. The restored Virtual Machine was not converted to a Template.\n";
	}

	my $vm_moref_find_view = Vim::find_entity_views(
					view_type => 'VirtualMachine',
					filter => { 'config.uuid' => $vm_uuid }
					);

	my $vm = shift @$vm_moref_find_view;
	eval {
		$vm->MarkAsTemplate();
	};
	if ($@) {
		# If VM conversion fails, log it, but don't fail, since it is not fatal
		my $failed_error = "Template conversion failed.\n";

		if (ref($@) eq 'SoapFault') {
			my $reason = "Error converting the restored Virtual Machine to a Template -";
			if (ref($@->detail) eq 'InvalidPowerState') {
				$$warning_ref = $$warning_ref . "$reason Virtual Machine was powered ON before the conversion could take place.\n";
			} elsif (ref($@->detail) eq 'InvalidState') {
				$$warning_ref = $$warning_ref . "$reason Virtual Machine configuration information is not available/accessible.\n".$@."\n";
			} elsif (ref($@->detail) eq 'NotSupported') {
				$$warning_ref = $$warning_ref . "$reason Virtual Machine as a Template is not supported.\n".$@."\n";
			} elsif (ref($@->detail) eq 'RuntimeFault') {
				$$warning_ref = $$warning_ref . "$reason Runtime Fault - Communication error has occurred.\n".$@."\n";
			} elsif (ref($@->detail) eq 'VmConfigFault') {
				$$warning_ref = $$warning_ref . "$reason Template is incompatible with the host.\n".$@."\n";
			} else {
				$$warning_ref = $$warning_ref . "$failed_error";
			}
		} else {
			$$warning_ref = $$warning_ref . "Error converting the Virtual Machine to a Template: $failed_error\n";
		}
	}
}

sub do_restore {
	my $datastore_path = shift;
	my $esxhost = shift;
	my $new_name = shift;
	my $metadata = shift;
	my $pool = shift;
	my $astemplate = shift;
	my $xml_filename = shift;
	my $xml_target = shift;
	my $xml_root = shift;

	my $warning = "";
	my $wir_mode = 0;

	check_for_maintenance_mode($esxhost);

	my $xml_doc = load_xml($xml_filename);

	my $new_vm_moref = create_vm($xml_doc, $datastore_path, $esxhost, $new_name, $metadata, $pool, $wir_mode);

	if ($astemplate eq "YES") {
		# Convert the VM to Template
		my $warning = "";
		my $get_view_vm = Vim::get_view(mo_ref => $new_vm_moref);
		my $new_vm_uuid = $get_view_vm->config->uuid;
		print STDERR "Converting VM with UUID ".$new_vm_uuid." to a Template\n";
		
		convert_to_template($new_vm_uuid, \$warning);

		if ($warning ne "") {
			# Put any warnings we generated into the XML.
			VProtect::output_xml_warning($xml_target, $warning);
		}
	}

	# pass this new vm_moref out in the XML output
	my $restore_result = VProtect::xml_get_or_create($xml_target, 'vprotect_restore');
	VProtect::append_to_xml($restore_result, 'vm_moref', $new_vm_moref->{'value'});

	# Append disk filenames to XML output
	append_disk_filenames($restore_result, $new_vm_moref, $xml_target);

	$xml_root->addChild($restore_result);
}

#####################
## Instant Recovery

sub parse_disk_map {
	# $filename contains a map such as
	#  2000:/backups/vmware_i_r/2000/disk
	#  2001:/backups/vmware_i_r/2001/disk

	# Return a hashref which contains a map like
	# {
	#   2000 => '2000/disk',
	#   2001 => '2001/disk'
	# }

	my $filename = shift;
	my $nfsroot = shift;
	
	my $diskmap = {};

	open(DISKMAP, '<', $filename) || die "Could not open disk map " . $filename . "\n";
	while (<DISKMAP>) {
		chomp;
		my ($key,$file) = split /:/,$_,2;
		$diskmap->{$key} = $file;
	}
	close DISKMAP;

	print STDERR "Disk map:\n" . Dumper($diskmap) . "\n";

	return $diskmap;
}

sub create_datastore {
	my $hostname = shift;
	my $remote_host = shift;
	my $remote_path = shift;
	my $datastore_name = shift;

	my $new_datastore_moref;

	eval {

	# Get a host view for the host
	my $host_view = Vim::find_entity_view(
		view_type => 'HostSystem',
		filter => { 'name' => $hostname },
		properties => [ 'configManager', 'runtime' ]
	);

	if (not defined $host_view) {
		die "Host $hostname not found while creating a datastore.\n";
	}

	# Find the datastore manager
	if (not defined $host_view->configManager) {
		die "configManager for $hostname not defined.\n";
	}
	if (not defined $host_view->configManager->datastoreSystem) {
		die "datastoreSystem for $hostname not defined.\n";
	}
	my $dm = $host_view->configManager->datastoreSystem;
	my $dm_view = Vim::get_view(mo_ref => $dm);

	my $spec = HostNasVolumeSpec->new(
		accessMode  => 'readWrite',
		localPath => $datastore_name,
		remoteHost => $remote_host,
		remotePath => $remote_path,
		type => 'nfs'
	);  # may need localPath

	$new_datastore_moref = $dm_view->CreateNasDatastore(
		spec => $spec
	);

	};
	if ($@) {
		if (ref($@) eq 'SoapFault') {
			if (ref($@->detail) eq 'AlreadyExists') {
				die "Error creating datastore - path already exists. (" . $@->detail->name . ")\n";
			} elsif (ref($@->detail) eq 'DuplicateName') {
				die "Error creating datastore - Duplicate Name: " . $@->fault_string . "\n";
			} elsif (ref($@->detail) eq 'HostConfigFault') {
				die "Error creating datastore - unable to mount NAS volume\n";
			} elsif (ref($@->detail) eq 'HostNotConnected') {
				die "Error creating datastore - The host is disconnected from vCenter.\n";
			} elsif (ref($@->detail) eq 'InvalidArgument') {
				die "Error creating datastore - Invalid Argument.\n";
			} elsif (ref($@->detail) eq 'NoGateway') {
				die "Error creating datastore - No VMkernel gateway is configured.\n";
			} elsif (ref($@->detail) eq 'NoPermission') {
				die "Error creating datastore - NoPermission -- " . $@->detail->privilegeId . " required.\n";
			} elsif (ref($@->detail) eq 'NoVirtualNic') {
				die "Error creating datastore - No VMkernel NIC is configured.\n";
			} elsif (ref($@->detail) eq 'PlatformConfigFault') {
				# This error will be recieved if vmkernel errors while trying to mount the share for some reason.
				die "Error creating datastore - PlatformConfigFault. VMkernel failed to mount the NFS share. Ensure that the NFS server is accessible and that the DPU's firewall is configured to allow NFS. By default, only 8 NFS datastores are allowed, see http://kb.vmware.com/kb/2002957.\n";
			}
		}
		print STDERR Dumper($@) . "\n\n";
		die "Error creating datastore: " . $@ . "\n";
	}

	print STDERR "create_datastore returning new moref: " . Dumper(%$new_datastore_moref) . "\n";

	return $new_datastore_moref;
}

sub get_ds_moref {
	my $datastore_name = shift;
	my $hostname = shift;

	my %datastore_info = get_datastore_info($datastore_name, $hostname);
	my $datastore_moref = $datastore_info{'mor'};
	if ($datastore_moref eq 0) {
		die "Error accessing datastore information.";
	}

	return $datastore_moref;
}

# Similar to getStatus sub-routine. However, this additionally checks to see if
# there is any question pending after a request to power ON the VM is sent
sub get_power_call_status {

	print STDERR "Entering get_power_status\n";

	my ($new_vm_moref,$taskRef,$message) = @_;

	my $task_view = Vim::get_view(mo_ref => $taskRef);

	my $taskinfo = $task_view->info->state->val;
	my $continue = 1;
	while ($continue) {

		# Get a new view to the VM, since the old view won't get the updated
		# information about the 'question'
		my $vm_view = Vim::get_view(mo_ref => $new_vm_moref);
		if (defined($vm_view->runtime->question)) {
			print STDERR "Some question is pending. So, calling AnswerVM\n";
			$vm_view->AnswerVM(questionId => $vm_view->runtime->question->id,
					answerChoice => $vm_view->runtime->question->choice->defaultIndex);
		} else {
			print STDERR "Currently no question is pending for this VM.\n";
		}

		my $info = $task_view->info;
		if ($info->state->val eq 'success') {
			print STDERR $message,"\n";
			$continue = 0;
		} elsif ($info->state->val eq 'error') {
			my $soap_fault = SoapFault->new;
			$soap_fault->name($info->error->fault);
			$soap_fault->detail($info->error->fault);
			$soap_fault->fault_string($info->error->localizedMessage);
			die "$soap_fault\n";
		}
		sleep 3;
		$task_view->ViewBase::update_view_data();
	}

	print STDERR "Leaving get_power_status\n";

}

sub power_on_vm {
	my $new_vm_moref = shift;

	my $vm_view = Vim::get_view(mo_ref => $new_vm_moref);

	my $task_ref;

	if ($vm_view->runtime->powerState ne "poweredOn") {
		eval {
			# Call the API with "_Task", so that if there is some question pending, it can be answered
			$task_ref = $vm_view->PowerOnVM_Task();
		};
		if ($@) {
			my $hdr = 'Failed to power on VM: ';
			if (ref($@) eq 'SoapFault') {
				if (ref($@->detail) eq 'CpuIncompatible1ECX') {
					die $hdr . $@->fault_string . "\n";
				}
				if (ref($@->detail) eq 'CpuIncompatible81EDX') {
					die $hdr . $@->fault_string . "\n";
				}
				if (ref($@->detail) eq 'DisallowedOperationOnFailoverHost') {
					die $hdr . "The host specified is a failover host.\n";
				}
				if (ref($@->detail) eq 'FileFault') {
					die $hdr . "There is a problem accessing the VM on the filesystem. (FileFault)\n";
				}
				if (ref($@->detail) eq 'GenericVmConfigFault') {
					die $hdr . "Generic VM Config Fault: " . $@->fault_string . "\n";
				}
				if (ref($@->detail) eq 'InsufficientResourcesFault') {
					die $hdr . "Doing so would violate a resource usage policy.\n";
				}
				if (ref($@->detail) eq 'InvalidPowerState') {
					die $hdr . "VM is already powered on.\n";
				}
				if (ref($@->detail) eq 'NoCompatibleHost') {
					die $hdr . "No host is compatible with the virtual machine.\n" . $@->detail;
				}
				if (ref($@->detail) eq 'NoPermission') {
					die $hdr . "No permission. Requires " . $@->detail->privilegeId . "\n";
				}
				if (ref($@->detail) eq 'NotEnoughLicenses') {
					die $hdr . "Not enough licenses to power on this VM.\n";
				}
				if (ref($@->detail) eq 'NotSupported') {
					die $hdr . "NotSupported -- VM is a template.\n";
				}
				if (ref($@->detail) eq 'RuntimeFault') {
					die $hdr . "RuntimeFault.\n" . Dumper($@) . "\n";
				}
				if (ref($@->detail) eq 'SystemError') {
					die $hdr . "SystemError: " . $@->detail->fault_string . "\n";
				}
				if (ref($@->detail) eq 'TaskInProgress') {
					die $hdr . "Task in progress -- VM is busy.\n";
				}
				if (ref($@->detail) eq 'VmConfigFault') {
					die $hdr . "Configuration issue.\n" . Dumper($@) . "\n";
				}
			}

			# We failed to identify the fault.
			die "Failed to power on VM.\n" . Dumper($@);
		}

		my $msg = "Successfully powered on the VM";
		get_power_call_status($new_vm_moref, $task_ref, $msg);
	}
}

sub power_off_vm {

	my $new_vm_moref = shift;

	my $vm_view = Vim::get_view(mo_ref => $new_vm_moref);

	print STDERR "Power State is - ".$vm_view->runtime->powerState->val."\n";
	my $task_ref;

	if ($vm_view->runtime->powerState->val ne "poweredOff") {
		eval {
			# Call the API with "_Task", so that if there is some question pending, it can be answered
			$task_ref = $vm_view->PowerOffVM_Task();
		};
		if ($@) {
			my $hdr = 'Failed to power OFF VM: ';
			if (ref($@) eq 'SoapFault') {
				if (ref($@->detail) eq 'InvalidPowerState') {
					die $hdr . "VM is already powered OFF.\n";
				}
				if (ref($@->detail) eq 'InvalidState') {
					die $hdr . "VM can't be powered OFF since the VM is in invalid state.\n";
				}
				if (ref($@->detail) eq 'NotSupported') {
					die $hdr . "This entity is a Template and not a VM. This operation can't be performed.\n";
				}
				if (ref($@->detail) eq 'RuntimeFault') {
					die $hdr . "RuntimeFault.\n" . Dumper($@) . "\n";
				}
				if (ref($@->detail) eq 'TaskInProgress') {
					die $hdr . "Task in progress -- VM is busy.\n";
				}
			}

			# We failed to identify the fault.
			die "Failed to power OFF VM.\n" . Dumper($@);
		}

		my $msg = "Successfully powered off the VM";
		get_power_call_status($new_vm_moref, $task_ref, $msg);
	}

}

sub migrate_vm ($$$$) {
	# Storage vMotion VM to final datastore

	my $new_vm_moref = shift;
	my $datastore_path = shift;
	my $hostname = shift;
	my $s_file = shift;   # Status file

	my $vm_view = Vim::get_view(mo_ref => $new_vm_moref);

	my $ds_moref = get_ds_moref($datastore_path, $hostname);
	#print "ds_moref is " . Dumper($ds_moref) . "\n";
	
	my @disk_locators;
	my $devices = $vm_view->config->hardware->device;
	foreach my $device (@$devices) {
		if (ref($device) eq 'VirtualDisk') {
			my $key = $device->key;
			my $dl = VirtualMachineRelocateSpecDiskLocator->new(
				datastore => $ds_moref,
				diskId => $key
			);

			push @disk_locators, $dl;
		}
	}

	#print "disk_locators is " . Dumper(@disk_locators) . "\n";
	#print "disk_locators[0] is " . Dumper($disk_locators[0]) . "\n";

	my $task_ref;

	eval {
		my $vm_reloc_spec = VirtualMachineRelocateSpec->new(
			datastore => $ds_moref,
			disk => \@disk_locators
		);
		
		#$vm_view->RelocateVM(spec => $vm_reloc_spec);
		#$task_ref = $vm_view->RelocateVM_Task(spec => $vm_reloc_spec, priority => 'highPriority');
		$task_ref = $vm_view->RelocateVM_Task(spec => $vm_reloc_spec);
	}; if ($@) {
		if (ref($@) eq 'SoapFault') {
			if (ref($@->detail) eq 'NoPermission') {
				die "RelocateVM failed: permission denied -- " . $@->detail->privilegeId . " permission is required in your vCenter role.\n";
			}
			if (ref($@->detail) eq 'NotSupported') {
				die "RelocateVM failed: The operation is not supported on the object.  Ensure that this is a vCenter server and not a standalone ESX host.\n";   # TODO ensure this message is right
			}
		}

		die "RelocateVM failed: " . Dumper($@) . "\n";
	}

	my $task = Vim::get_view(mo_ref => $task_ref);

	sleep 2;
	
	while (($task->info->state->val eq 'running') or
		($task->info->state->val eq 'queued')) {
		sleep 15;
		my $reloc_percent_done = $task->info->progress;
		# We are reporting percentages between 25 and 90% -- scale to that range.
		$reloc_percent_done *= 0.65;
		$reloc_percent_done += 25;
		$reloc_percent_done = int($reloc_percent_done);
		update_status($s_file, 'available', $reloc_percent_done, 'Relocating VM');

		$task->update_view_data();
	}

	if ($task->info->state->val eq 'success') {
		print STDERR "Relocation task succeeded.\n";
	} elsif ($task->info->state->val eq 'error') {
		print STDERR "Relocation task failed.\n";
		print STDERR Dumper($task);
		my $error_msg = '<no further information provided>';
		if (defined($task->info->error->localizedMessage)) {
			$error_msg = $task->info->error->localizedMessage;
		}
		die "Could not relocate VM: $error_msg\n";
	} else {
		die "Unknown state during RelocateVM task: " . $task->info->state->val . "\n" . Dumper($task->info) . "\n";
	}
}

sub remove_datastore ($$) {
	my $hostname = shift;
	my $ds_moref = shift;

	# Get a host view for the host
	my $host_view = Vim::find_entity_view(
		view_type => 'HostSystem',
		filter => { 'name' => $hostname },
		properties => [ 'configManager', 'runtime' ]
	);

	if (not defined $host_view) {
		die "Host $hostname not found while removing a datastore.\n";
	}

	eval {

	# Find the datastore manager
	if (not defined $host_view->configManager) {
		die "configManager for $hostname not defined while removing datastore.\n";
	}
	if (not defined $host_view->configManager->datastoreSystem) {
		die "datastoreSystem for $hostname not defined while removing datastore.\n";
	}
	my $dm = $host_view->configManager->datastoreSystem;
	my $dm_view = Vim::get_view(mo_ref => $dm);

	$dm_view->RemoveDatastore(datastore => $ds_moref);
	
	};
	if ($@) {
		my $hdr = "Error removing datastore -";
		my $msg;
		if (ref($@) eq 'SoapFault') {
			if (ref($@->detail) eq 'HostConfigFault') {
				die "$hdr HostConfigFault\n";
			} elsif (ref($@->detail) eq 'NoPermission') {
				die "$hdr NoPermission -- requires " . $@->detail->privilegeId . " permission in vCenter.\n";
			} elsif (ref($@->detail) eq 'NotFound') {
				die "$hdr NotFound.\n";
			} elsif (ref($@->detail) eq 'ResourceInUse') {
				die "$hdr " . $@->detail->name . ": Resource In Use. VMs must be removed from this datastore before it can be removed.\n";
			} elsif (ref($@->detail) eq 'RuntimeFault') {
				die "$hdr RuntimeFault.\n";
			}
		} else {
			die "$hdr unknown error.\n" . Dumper($@) . "\n";
		}
	}
}

sub create_ir_snapshot ($) {
	my $vm_moref = shift;

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);

	my $snap_moref;
	my $snap_name = "Unitrends Instant Recovery";
	my $description = "Instant recovery auditing mode";

	eval {
		$snap_moref = $vm_view->CreateSnapshot(
			name => 'Unitrends Instant Recovery',
			description => 'Instant recovery auditing mode',
			memory => 0,
			quiesce => 0
		);
	};
	if ($@) {
		# If this failed, log a note, but do not fail IR.
			print STDERR "Unitrends IR snapshot failed.\n" . Dumper($@) . "\n";
	}

	return $snap_moref;
}

sub disconnect_network ($) {
	# For each NIC in the VM, set it to start disconnected.

	my $vm_moref = shift;
	
	my $conn = VirtualDeviceConnectInfo->new(
		allowGuestControl => 1,
		connected => 0,
		startConnected => 0
	);

	# For each network card
	my $vm_view = Vim::get_view(mo_ref => $vm_moref);
	my $devices = $vm_view->config->hardware->device;

	my $dc_array_ref = [];

	foreach (@$devices) {
		if ($_->isa('VirtualEthernetCard')) {
			my $nic_type = ref($_);

			my %hopt = ( connectable => $conn, key => $_->key );

			my $dev;
			if ($nic_type eq 'VirtualE1000') {
				$dev = VirtualE1000->new(%hopt);
			} elsif ($nic_type eq 'VirtualE1000e') {
				$dev = VirtualE1000e->new(%hopt);
			} elsif ($nic_type eq 'VirtualPCNet32') {
				$dev = VirtualPCNet32->new(%hopt);
			} elsif ($nic_type eq 'VirtualVmxnet') {
				$dev = VirtualVmxnet->new(%hopt);
			} elsif ($nic_type eq 'VirtualVmxnet2') {
				$dev = VirtualVmxnet2->new(%hopt);
			} elsif ($nic_type eq 'VirtualVmxnet3') {
				$dev = VirtualVmxnet3->new(%hopt);
			} elsif ($nic_type eq 'VirtualSriovEthernetCard') {
				$dev = VirtualSriovEthernetCard->new(%hopt);
			} else {
				die "Error disconnecting NIC: unrecognized type: " . $nic_type . "\n";
			}
			
			my $spec = VirtualDeviceConfigSpec->new(
				device => $dev,
				operation => VirtualDeviceConfigSpecOperation->new('edit')
			);

			push @$dc_array_ref, $spec;
		}
	}

	if (scalar @$dc_array_ref > 0) {
		# We are changing at least one NIC.
		my $change_spec = VirtualMachineConfigSpec->new(
			deviceChange => $dc_array_ref
		);

		my $task_ref;
		my $retry = 1;
		print STDERR "Before ReconfigVM_Task for disconnecting the network\n";

		eval {
			$task_ref = $vm_view->ReconfigVM_Task(spec => $change_spec);
		};
		if ($@) {
			if (handle_reconfigvm_error($retry, \$@) == 1) {
				# handle_reconfigvm_error will return 1, when ReconfigVM_Task API
				# fails to reconfigure the VM due to "TaskInProgress" error.
				# Worth to try the reconfigure operation again.
				eval {
					$task_ref = $vm_view->ReconfigVM_Task(spec => $change_spec);
				};
				if ($@) {
					$retry = 0;
					handle_reconfigvm_error($retry, \$@);
				}
			}
		}

		my $msg = "Successfully disconnected the VM network";
		VProtect::getStatus($task_ref, $msg);

		print STDERR "Reconfigure VM task to disconnect the VM network was successful\n";
	}
}

sub connect_network {

	# For each NIC in the VM, set it to start disconnected.

	my $vm_moref = shift;

	my $conn = VirtualDeviceConnectInfo->new(
		allowGuestControl => 1,
		connected => 1,
		startConnected => 1
	);

	# For each network card
	my $vm_view = Vim::get_view(mo_ref => $vm_moref);
	my $devices = $vm_view->config->hardware->device;

	my $dc_array_ref = [];

	foreach (@$devices) {
		if ($_->isa('VirtualEthernetCard')) {
			my $nic_type = ref($_);

			my %hopt = ( connectable => $conn, key => $_->key );

			my $dev;
			if ($nic_type eq 'VirtualE1000') {
				$dev = VirtualE1000->new(%hopt);
			} elsif ($nic_type eq 'VirtualE1000e') {
				$dev = VirtualE1000e->new(%hopt);
			} elsif ($nic_type eq 'VirtualPCNet32') {
				$dev = VirtualPCNet32->new(%hopt);
			} elsif ($nic_type eq 'VirtualVmxnet') {
				$dev = VirtualVmxnet->new(%hopt);
			} elsif ($nic_type eq 'VirtualVmxnet2') {
				$dev = VirtualVmxnet2->new(%hopt);
			} elsif ($nic_type eq 'VirtualVmxnet3') {
				$dev = VirtualVmxnet3->new(%hopt);
			}  elsif ($nic_type eq 'VirtualSriovEthernetCard') {
				$dev = VirtualSriovEthernetCard->new(%hopt)
			} else {
				die "Error connecting NIC: unrecognized type: " . $nic_type . "\n";
			}

			my $spec = VirtualDeviceConfigSpec->new(
				device => $dev,
				operation => VirtualDeviceConfigSpecOperation->new('edit')
			);

			push @$dc_array_ref, $spec;
		}
	}

	if (scalar @$dc_array_ref > 0) {
		# We are changing at least one NIC.
		my $change_spec = VirtualMachineConfigSpec->new(
			deviceChange => $dc_array_ref
		);

		my $task_ref;
		my $retry = 1;
		print STDERR "Before ReconfigVM_Task for connecting the network\n";

		eval {
			$task_ref = $vm_view->ReconfigVM_Task(spec => $change_spec);
		};
		if ($@) {
			if (handle_reconfigvm_error($retry, \$@) == 1) {
				# handle_reconfigvm_error will return 1, when ReconfigVM_Task API
				# fails to reconfigure the VM due to "TaskInProgress" error.
				# Worth to try the reconfigure operation again.
				eval {
					$task_ref = $vm_view->ReconfigVM_Task(spec => $change_spec);
				};
				if ($@) {
					$retry = 0;
					handle_reconfigvm_error($retry, \$@);
				}
			}
		}

		my $msg = "Successfully connected the VM network";
		VProtect::getStatus($task_ref, $msg);

		print STDERR "Reconfigure VM task to connect the VM network was successful\n";
	}

}

sub update_status ($$$$) {
	my $s_file = shift;
	my $state = shift;
	my $percent = shift;
	my $detail = shift;

	die if (not defined($detail));

	# File has a line in it like
	# status:running:50
	# which indicates 50% done.  Update this percentage value.

	if (defined($s_file)) {
		`sed -i s/^status:.*/status:$state:$percent:"$detail"/ $s_file 2>/dev/null`;
	}
}

sub record_datastore_info ($$$$$$$) {
	my $s_file = shift;
	my $datastore_moref = shift;
	my $datastore_name = shift;
	my $server = shift;
	my $hostname = shift;
	my $username = shift;
	my $password = shift;

	my $password_b64 = encode_base64($password);

	my $xmloutput = XML::LibXML->createDocument;
	$xmloutput->setEncoding('UTF-8');	
	my $xmlroot = $xmloutput->createElement('datastore_info');
	$xmloutput->setDocumentElement($xmlroot);

	$xmlroot->appendTextChild('datastore_moref', $datastore_moref->{'value'});
	$xmlroot->appendTextChild('datastore_name',  $datastore_name);
	$xmlroot->appendTextChild('server',          $server);
	$xmlroot->appendTextChild('hostname',        $hostname);
	$xmlroot->appendTextChild('username',        $username);
	$xmlroot->appendTextChild('password',        $password_b64);

	open(STATUS, '>>', $s_file . '.ds_info') || die "Cannot open $s_file in append mode: $!\n";
	print STATUS $xmloutput->toString(1);
	close STATUS;
}

sub record_vm_info {

	my $s_file = shift;
	my $vm_uuid = shift;
	my $vm_name = shift;
	my $server = shift;
	my $hostname = shift;   # May require it later
	my $username = shift;
	my $password = shift;
	my $ir_datastore_moref = shift;
	my $ir_datastore_name = shift;
	my $ir_audit = shift;
	my $vm_moref = shift;

	my $password_b64 = encode_base64($password);

	my $xmloutput = XML::LibXML->createDocument;
	$xmloutput->setEncoding('UTF-8');
	my $xmlroot = $xmloutput->createElement('vm_info');
	$xmloutput->setDocumentElement($xmlroot);

	$xmlroot->appendTextChild('vm_uuid', $vm_uuid);
	$xmlroot->appendTextChild('vm_name', $vm_name);
	$xmlroot->appendTextChild('server', $server);
	$xmlroot->appendTextChild('hostname', $hostname);
	$xmlroot->appendTextChild('username', $username);
	$xmlroot->appendTextChild('password', $password_b64);
	$xmlroot->appendTextChild('ir_datastore_moref', $ir_datastore_moref->{'value'});
	$xmlroot->appendTextChild('ir_datastore_name', $ir_datastore_name);
	$xmlroot->appendTextChild('audit', $ir_audit);
	$xmlroot->appendTextChild('vm_moref', $vm_moref);

	open(STATUS, '>>', $s_file . '.vm_info') || die "Cannot open $s_file in append mode: $!\n";
	print STATUS $xmloutput->toString(1);
	close STATUS;

}

sub check_esx_version_for_instant ($) {
	my $hostname = shift;
	
	my $host_view = Vim::find_entity_view(
		view_type => 'HostSystem',
		filter => { 'name' => $hostname },
		properties => [ 'config' ]
	);

	if (not defined $host_view) {
		die "ESX Host $hostname not found while checking version.\n";
	}

	if ($host_view->config->product->version =~ '^3\.') {
		die "Instant recovery is only supported on ESX/ESXi 4+.\n";
	}
}

sub check_vcenter_version_for_instant {
	my $content = vim_get_service_content();
	my $about = $content->about;

	if ($about->{'apiType'} ne 'VirtualCenter') {
		die "Instant recovery attempted on a non-vCenter server.\n";
	}

	my $version = $about->{'version'};
	my $major = (split /\./,$version,1)[0];
	if ($major < 4) {
		die "Instant recovery is only supported on vCenter Server 4 and newer.\n";
	}
}

sub get_vm_info_for_ir {
	my $new_vm_moref = shift;

	my $get_view_vm = Vim::get_view(mo_ref => $new_vm_moref);
	my $new_vm_uuid = $get_view_vm->config->uuid;
	my $new_vm_name = $get_view_vm->config->name;

	my $new_vm_moref_string = $new_vm_moref->value;

	print STDERR "VM name is  - $new_vm_name\n";
	print STDERR "VM UUID is  - $new_vm_uuid\n";
	print STDERR "VM MoRef is - $new_vm_moref_string\n";

	return($new_vm_uuid, $new_vm_name, $new_vm_moref_string);

}

sub do_instant (%) {
	my %arg_hash = @_;
	my $hostname          = $arg_hash{'hostname'};
	my $pool              = $arg_hash{'pool'};
	my $newvmname         = $arg_hash{'newvmname'};
	my $datastore_path    = $arg_hash{'datastore_name'};
	my $ir_datastore_host = $arg_hash{'ir_datastore_host'};
	my $ir_datastore_path = $arg_hash{'ir_datastore_path'};
	my $ir_disk_map_path  = $arg_hash{'ir_disk_map_path'};
	my $xml_filename      = $arg_hash{'xml_filename'};
	my $s_file            = $arg_hash{'ir_status_file'};
	my $power	      = $arg_hash{'power'};
	my $uuid              = $arg_hash{'uuid'};      # ESX server's UUID
	my $server            = $arg_hash{'server'};
	my $username          = $arg_hash{'username'};
	my $password          = $arg_hash{'password'};

	my $ir_audit = 0;
	if (defined($arg_hash{'ir_audit'})) {
		$ir_audit = 1;
	}

	my $wir_mode = 0;

	check_esx_version_for_instant($hostname);

	check_for_maintenance_mode($hostname);

	my $xml_doc = load_xml($xml_filename);

	my $disk_map_href = parse_disk_map($ir_disk_map_path, $ir_datastore_path);

	# "ir_datastore_path" is of the form "/backups/vmware_ir/<esx_uuid>/restore_xxx"
	my @disk_location = split(/\//, $ir_datastore_path);
	print STDERR "IR Datastore path information -\n" . Dumper(@disk_location) . "\n";

	my $datastore_name;
	my $ir_ds_exists;
	my $ir_datastore_moref;
	my $nas_ds_name_prefix = "Unitrends_Instant_Recovery";
	($ir_ds_exists, $datastore_name, $ir_datastore_moref) = check_nas_ds_exists($hostname, $ir_datastore_host, $nas_ds_name_prefix);
	print STDERR "ir_ds_exists is $ir_ds_exists\ndatastore_name is $datastore_name\nir_datastore_moref is $ir_datastore_moref\n";

	if ($ir_ds_exists) {
		print STDERR "Datastore for VMware IR already exists on this ESX server.\n";
	} else {
		update_status($s_file, 'running', 10, 'Creating datastore');
		$datastore_name = "Unitrends_Instant_Recovery_$$";
		my $ir_datastore_path_temp = "/backups/vmware_ir/$uuid/";
		$ir_datastore_moref = create_datastore($hostname, $ir_datastore_host, $ir_datastore_path_temp, $datastore_name);
		print STDERR "MoRef is $ir_datastore_moref\n";
		# In some cases, a 'Rename Datastore' command is issued from 'somewhere',
		# which changes the name of the NAS Datastore
		sleep 20;
	}

	my $ds_view = Vim::get_view(mo_ref => $ir_datastore_moref);
	print STDERR "Name of DS from MoRef is " .$ds_view->name. "\n";
	$datastore_name = $ds_view->name;
	
	update_status($s_file, 'running', 15, 'Creating VM');

	record_datastore_info($s_file, $ir_datastore_moref, $datastore_name, $server, $hostname, $username, $password);

	#print "New datastore: " . Dumper($ir_datastore_moref) . "\n";
	my $metadata = "NO";	# Do not Restore metadata when performing Instant Recovery
	my $new_vm_moref = create_vm($xml_doc, $datastore_path, $hostname, $newvmname, $metadata, $pool, $wir_mode, { ir_datastore_moref => \$ir_datastore_moref, datastore_name => $datastore_name, disk_map_href => $disk_map_href, restore_dir => $disk_location[4], ir_audit => $ir_audit });

	# Get the VM's name and UUID
	my ($vm_uuid, $vm_name, $vm_moref_string) = get_vm_info_for_ir($new_vm_moref);
	record_vm_info($s_file, $vm_uuid, $vm_name, $server, $hostname, $username, $password, $ir_datastore_moref, $datastore_name, $ir_audit, $vm_moref_string);

	if ($ir_audit) {
		# Auditing mode

		update_status($s_file, 'running', 30, 'Creating snapshot');

		my $snap_moref = create_ir_snapshot($new_vm_moref);

		update_status($s_file, 'running', 35, 'Configuring network adapters');

		disconnect_network($new_vm_moref);

		update_status($s_file, 'running', 50, 'Powering on VM');

		if ($power) {
			power_on_vm($new_vm_moref);
		}
		
		update_status($s_file, 'available', 90, 'Cleaning up');
	} else {
		update_status($s_file, 'running', 20, 'Powering on VM');

		power_on_vm($new_vm_moref);

		update_status($s_file, 'available', 24, 'Waiting for VM');
		sleep(90);

		update_status($s_file, 'available', 25, 'Relocating VM');

		migrate_vm($new_vm_moref, $datastore_path, $hostname, $s_file);

		#update_status($s_file, 'available', 90, 'Removing datastore');

		#remove_datastore($hostname, $ir_datastore_moref);

		#update_status($s_file, 'available', 95, 'Cleaning up');
	}
}

sub ensure_ds_empty {
	my $hostname = shift;
	my $ds_moref = shift;

	my $ds_view = Vim::get_view(mo_ref => $ds_moref);

	my $vms = Vim::get_views(
			mo_ref_array => $ds_view->vm,
			properties => ['name']
		);

	my $number_of_vms = scalar @$vms;

	if ($number_of_vms > 0) {
		Util::disconnect();
		exit(33);
	}
}

sub ensure_datastore_empty {
	my %arg_hash = @_;
	my $hostname = $arg_hash{'hostname'};
	my $datastore_moref_value = $arg_hash{'datastore_moref_value'};
	my $datastore_name = $arg_hash{'datastore_name'};

	my $ds_moref = bless(
		{ value => $datastore_moref_value, type => 'Datastore'},
		'ManagedObjectReference');

	ensure_ds_empty($hostname, $ds_moref);
}


sub remove_ir_datastore {
	my %arg_hash = @_;
	my $hostname = $arg_hash{'hostname'};
	my $datastore_moref_value = $arg_hash{'datastore_moref_value'};
	my $datastore_name = $arg_hash{'datastore_name'};

	my $ds_moref = bless(
			{ value => $datastore_moref_value, type => 'Datastore'},
			'ManagedObjectReference');

	# Check that it has the expected name
	my $ds_view = Vim::get_view(mo_ref => $ds_moref);
	if ($ds_view->name ne $datastore_name) {
		die "Datastore name " . $ds_view->name . " does not match expected name $datastore_name, refusing to remove datastore.\n";
	}

	remove_datastore($hostname, $ds_moref);
}

sub ensure_vm_removed {

	my $vmuuid = shift;
	my $ir_ds_name = shift;
	my $ir_ds_moref = shift;

	my $service_content = Vim::get_service_content();
	my $search_index = Vim::get_view(mo_ref => $service_content->searchIndex);
	my $vm_moref = $search_index->FindByUuid(
		uuid => $vmuuid,
		vmSearch => 1
	);

	if (defined $vm_moref) {

		# VM exists in the inventory. Now check if it's disks still belong to the
		# Unitrends_Instant_Recovery_xxx datastore
		# This is the case -
		#       1. When "RDR" brings up the VM in Audit Mode and performs
		#          a Storage vMotion later (called as FAILOVER)
		# OR    2. When a user brings up the VM in Audit Mode and later performs
		#          the Storage vMotion manually (outside of our knowledge)

		my $get_view_vm = Vim::get_view(mo_ref => $vm_moref);
		my $new_vm_uuid = $get_view_vm->config->uuid;
		my $new_vm_name = $get_view_vm->config->name;
		print STDERR "VM name is $new_vm_name\n";

		my $devices = $get_view_vm->config->hardware->device;
		foreach (@$devices) {
			if (ref($_) eq 'VirtualDisk') {
				my $ds_view = Vim::get_view( mo_ref => $_->{'backing'}->datastore );
				my $ds_name = $ds_view->get_property('summary')->name;
				print STDERR "Datastore name on which disk with key " . $_->{'key'} .
				" resides is " . $ds_name . "\n";
				if ($ds_name =~ '^Unitrends_Instant_Recovery') {

					# TODO If required, also compare the mo-refs of the datastore

					# This means that at least one of the VM's disks still exists on the
					# Unitrends_Instant_Recovery Datastore. Can't delete the VM from the
					# Unitrends inventory
					print STDERR "The VM $new_vm_name with UUID $new_vm_uuid still has it's" .
					" disks on the Unitrends Instant Recovery datastore.\n";
					Util::disconnect();
					exit(34);
				}
			}
		}

		# If we reach here, it means the VM exists in the inventory, but none of it's disks are
		# on the Unitrends_Instant_Recovery datastore. Good.
		print STDERR "The VM $new_vm_name with UUID $new_vm_uuid still exists on the ESX/vCenter server.\n";
		print STDERR "However, it's disks have been Storage vMotion-ed. Safe to remove the information for" .
		" this VM from the Unitrends appliance\n";
	} else {
		# VM currently doesn't exist in the inventory. Good.
		print STDERR "VM with UUID $vmuuid doesn't exist in the inventory.\n";
	}

}

sub remove_all_snapshots {

	my $vm_view = shift;

	my $task_ref;
	eval {
		$task_ref = $vm_view->RemoveAllSnapshots_Task();
	};

	if ($@) {
		print STDERR "Dumping contents of error for removing all snapshots -\n" .Dumper($@). "\n";
		my $default_err = 'Failed to remove all snapshots: ';
		if (ref($@) eq 'SoapFault') {
			if (ref($@->detail) eq 'InvalidPowerState') {
				die $default_err . "Cannot remove snapshots in this power state\n";
			}
			if (ref($@->detail) eq 'InvalidState') {
				die $default_err . "Cannot access VM configuration to perform operation\n";
			}
			if (ref($@->detail) eq 'NotSupported') {
				die $default_err . "Host doesn't support this operation\n";
			}
			if (ref($@->detail) eq 'RuntimeFault') {
				die $default_err . "Communication error occurred\n";
			}
			if (ref($@->detail) eq 'SnapshotFault') {
				die $default_err . "Snapshot operation related error occurred\n";
			}
			if (ref($@->detail) eq 'TaskInProgress') {
				die $default_err . "VM is busy\n";
			}
		} else {
			die "Removing all snapshots failed\n";
		}
	}
	my $msg = "Successfully removed all snapshots";
	VProtect::getStatus($task_ref, $msg);

}

sub delete_vmir_vm {

	my $vmuuid = shift;
	my $xmldoc = shift;
	my $xmlroot = shift;

	my $vm_moref = VProtect::vim_get_vm_by_uuid($vmuuid);
	if (!defined $vm_moref) {
		# VM is not present in the inventory.
		# Don't error out. Jus return
		return;
	}

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);

	# For VMIR in Audit Mode, we create a 'Unitrends Instant Recovery' snapshot.
	# Before deleting the VM, revert to it and remove ALL snapshots
	my $snapshot_name = "Unitrends Instant Recovery";
	if (defined $vm_view->snapshot) {
		if (defined (@{$vm_view->snapshot->rootSnapshotList})) {
			foreach (@{$vm_view->snapshot->rootSnapshotList}) {
				revert_snap($_, $snapshot_name);
			}
		}
	} else {
		print STDERR "No VMIR snapshot found on the VM. Proceeding with deletion.";
	}

	# We are deleting the VM anyhow. So, safe to delete ALL the snapshots
	remove_all_snapshots($vm_view);

	# Now delete the VM
	# Just in case if the VM got powered ON, power it OFF first
	if ($vm_view->runtime->powerState->val eq "poweredOn") {
		power_off_vm($vm_moref);
	}

	my $task_ref;
	eval {
		$task_ref = $vm_view->Destroy_Task();
	};
	if ($@) {
		my $hdr = 'Failed to Delete the VM: ';
		if (ref($@) eq 'SoapFault') {
			if (ref($@->detail) eq 'RuntimeFault') {
				die $hdr . "RuntimeFault.\n" . Dumper($@) . "\n";
			}
			if (ref($@->detail) eq 'VimFault') {
				die $hdr . "VimFault.\n" . Dumper($@) . "\n";
			}
		}

		# Failed to identify the fault.
		die "Failed to delete the VM.\n" . Dumper($@);
	}

	my $msg = "Successfully deleted the VM";
	VProtect::getStatus($task_ref, $msg);

}

#####################################
## Windows Instant Recovery on VMware
#####################################

# Sub-routine to boot a VM always from CD-ROM
sub change_boot_order {

	print STDERR "Entering change_boot_order\n";

	my $vm_moref = shift;
	my $xmldoc = shift;
	my $xmlroot = shift;

	my $get_view_vm = Vim::get_view(mo_ref => $vm_moref);
	my $vm_uuid = $get_view_vm->config->uuid;

	my $vm = $xmldoc->createElement('virtual_machine');
		#append_if_def($vm, 'uuid', $vm_uuid);
	$vm->appendTextChild('uuid', $vm_uuid);
	$xmlroot->appendChild($vm);

	# Safety check
	my $new_vm_moref = VProtect::vim_get_vm_by_uuid($vm_uuid);
	if (!defined $new_vm_moref) {
		die "VM with UUID ".$vm_uuid." was not found.\n";
	}

	my $vm_moref_find_view = Vim::find_entity_view(
					view_type => 'VirtualMachine',
					filter => { 'config.uuid' => $vm_uuid }
					);

	my @bootOptions = ();
	my $tmpBootDev = '';
	$tmpBootDev = VirtualMachineBootOptionsBootableCdromDevice->new();
	push @bootOptions, $tmpBootDev;
	my $bootOptions = VirtualMachineBootOptions->new(bootOrder => \@bootOptions);
	my $spec = VirtualMachineConfigSpec->new(bootOptions => $bootOptions);

	# TODO Put the below set of code in a different independent sub-routine
	my $task_ref;
	my $retry = 1;
	print STDERR "Before ReconfigVM_Task for changing the boot order\n";

	eval {
		$task_ref = $vm_moref_find_view->ReconfigVM_Task(spec => $spec);
	};
	if ($@) {
		if (handle_reconfigvm_error($retry, \$@) == 1) {
			# handle_reconfigvm_error will return 1, when ReconfigVM_Task API
			# fails to reconfigure the VM due to "TaskInProgress" error.
			# Worth to try the reconfigure operation again.
			eval {
				$task_ref = $vm_moref_find_view->ReconfigVM_Task(spec => $spec);
			};
			if ($@) {
				$retry = 0;
				handle_reconfigvm_error($retry, \$@);
			}
		}
	}

	my $msg = "Successfully updated VM Boot Options";
	VProtect::getStatus($task_ref, $msg);

	print STDERR "Reconfigure VM task to change the boot order was successful\n";

	print STDERR "Leaving change_boot_order\n";
}

# Sub-routine to check if the 'switch' is a Distributed
# Virtual Switch or not
sub check_network_setting {

	print STDERR "Entering check_network_setting\n";

	my ($vm_moref, $hostname, $switch, $xmldoc, $xmlroot) = @_;

	# Get a host view for the host
	my $host_view = Vim::find_entity_view(
			view_type => 'HostSystem',
			filter => { 'name' => $hostname },
			properties => [ 'network' ]
			);

	if (not defined $host_view) {
		return 0;	# Won't happen, but just in case
	}

	my $nw_view;
	my $dvs_view;
	my $dvs_found = 0;
	my $host_network = $host_view->network;
	foreach(@$host_network) {
		if ($_->type eq 'DistributedVirtualPortgroup') {
			$nw_view = Vim::get_view( mo_ref => $_,
						properties => [ 'summary.accessible',
								'summary.name',
								'config.distributedVirtualSwitch' ]);

			$dvs_view = Vim::get_view( mo_ref => $nw_view->{'config.distributedVirtualSwitch'},
						properties => [ 'name' ] );

			print STDERR "\t N/W name / DVPG Name is - " . $nw_view->{'summary.name'} . "\n";
			print STDERR "\t DVS name                - " . $dvs_view->{'name'} . "\n";

			if ($nw_view->{'summary.accessible'} eq 'true') {
				if( defined($dvs_view) ) {
					my $dv_name = $nw_view->{'summary.name'} . " (" . $dvs_view->{'name'} . ")";
					if ($dv_name eq $switch) {
						# Found the required vDS. Good.
						# Now point the NIC to appropriate network
						change_network_setting($vm_moref, $nw_view->{'summary.name'}, $switch, $xmldoc, $xmlroot);
						$dvs_found = 1;
						last;
					}
				}
			}
		}
	}

	print STDERR "Leaving check_network_setting\n";
	if ($dvs_found) {
		return 1;
	} else {
		return 0;
	}

}

# This will only be called for a Distributed Virtual Switch
sub change_network_setting {

	my ($vm_moref, $nw_view_name, $switch, $xmldoc, $xmlroot) = @_;

	my $vm_get_view = Vim::get_view(mo_ref => $vm_moref);   # Used later for reconfiguration of the VM

	# DVPG = Distributed Virtual Port Group
	my $dvpg_view = Vim::find_entity_view( view_type => 'DistributedVirtualPortgroup',
						filter => { 'name' => $nw_view_name });
	my $pg_key = $dvpg_view->key;
	my $dvs_key = $dvpg_view->config->distributedVirtualSwitch;
	my $dvs_entity = Vim::get_view(mo_ref => $dvs_key);
	my $dvs_uuid = $dvs_entity->uuid;

	my $dvs_port_conn = DistributedVirtualSwitchPortConnection->new(
					portgroupKey => $pg_key,
					switchUuid => $dvs_uuid);

	my $virtual_dev_backing_info = VirtualEthernetCardDistributedVirtualPortBackingInfo->new(
					port => $dvs_port_conn);

	my $virtual_dev_connect_info = VirtualDeviceConnectInfo->new(
					startConnected => 1,
					allowGuestControl => 1,
					connected => 1);

	# New object will be "VirtualE1000" since, we created it
	# Key for the same is 4000
	my $network_adapter = VirtualE1000->new( key => 4000,
						backing => $virtual_dev_backing_info,
						connectable => $virtual_dev_connect_info );

	my @spec = VirtualDeviceConfigSpec->new( device => $network_adapter,
						operation => VirtualDeviceConfigSpecOperation->new('edit') );

	my $change_spec = VirtualMachineConfigSpec->new( deviceChange => \@spec );

	my $retry = 1;
	my $task_ref;
	eval {
		$task_ref = $vm_get_view->ReconfigVM_Task(spec => $change_spec);
	};
	if ($@) {
		if (handle_reconfigvm_error($retry, \$@) == 1) {
			# handle_reconfigvm_error will return 1, when ReconfigVM_Task API
			# fails to reconfigure the VM due to "TaskInProgress" error.
			# Worth to try the reconfigure operation again.
			eval {
				$task_ref = $vm_get_view->ReconfigVM_Task(spec => $change_spec);
			};
			if ($@) {
				$retry = 0;
				handle_reconfigvm_error($retry, \$@);
			}
		}
	}

	my $msg = "Successfully connected to a DVS";
	VProtect::getStatus($task_ref, $msg);

}

sub handle_reconfigvm_error {

	my $retry = shift;      # 1 = return 1 so that a retry can be done, 0 = die
	my (@error) = shift;    # The error is passed by reference

	my $default_error = "Error reconfiguring the VM";

	if (ref($@) eq 'SoapFault') {
		if (ref($@->detail) eq 'ConcurrentAccess') {
			die "$default_error: The changeVersion of the virtual machine doesn't match the changeVersion of the ESX " .
				"server for the configuration.\n";
		}
		if (ref($@->detail) eq 'DuplicateName') {
			die "$default_error: Cannot change the name of the virtual machine, since this name already exists.\n";
		}
		if (ref($@->detail) eq 'FileFault') {
			die "$default_error: Cannot access the virtual machine's files for this operation.\n";
		}
		if (ref($@->detail) eq 'InsufficientResourcesFault') {
			die "$default_error: The reconfiguration operation violates a resource usage policy. Place the virtual machine " .
				"where enough resources are available.\n";
		}
		if (ref($@->detail) eq 'InvalidName') {
			die "$default_error: The specified name  of the virtual machine is invalid.\n";
		}
		if (ref($@->detail) eq 'InvalidPowerState') {
			die "$default_error: The virtual hardware cannot support the configuration changes. So, cannot power ON the virtual machine.\n";
		}
		if (ref($@->detail) eq 'InvalidState') {
			die "$default_error: The virtual machine's configuration is not accessible.\n";
		}
		if (ref($@->detail) eq 'RuntimeFault') {
			die "$default_error: A communication error occurred.\n";
		}
		if (ref($@->detail) eq 'TaskInProgress') {
			if ($retry) {
				# This fault is returned sometimes, when there is already some task pending on the VM.
				# This can be a legitimate fault. So, sleep for a short period.
				sleep 30;
				print STDERR "Reconfiguring the virtual machine was unsuccessful for the first time. Trying it again.\n";
				return 1;
			} else {
				die "$default_error: A task is already in progress.\n";
			}
		}
		if (ref($@->detail) eq 'TooManyDevices') {
			die "$default_error: The device specifications exceed the allowed limits.\n";
		}
		if (ref($@->detail) eq 'VmConfigFault') {
			# Thrown if the specified spec for reconfiguration is invalid
			die "$default_error: VmConfigFault\n" . $@ . "\n";
		}
		if (ref($@->detail) eq 'InvalidDeviceOperation') {
			my $badOp = "";
			if (defined($@->detail->badOp)) {
				$badOp = $@->detail->badOp->val . " on deviceIndex " . $@->detail->deviceIndex . "\n";
			}
			die "$default_error: Invalid device operation - " . $badOp . "\n" . Dumper($@->detail);
		}
		if (ref($@->detail) eq 'NoPermission') {
			die "$default_error: Requires permission " . $@->detail->privilegeId . "\n";
		}
	}

	# Failure was not caught above or was not a SoapFault
	die "$default_error: " . $@ . "\n";

}

# Sub-routine to check whether an WIR Datastore with the name
# "Unitrends_WIR_xxxxx" exists or not on the ESX Server '$hostname'
sub check_nas_ds_exists {

	my $hostname = shift;
	my $nas_ds_host = shift;
	my $nas_ds_name_prefix = shift;

	my $ds_name;
	my $ds_moref;
	my $nas_ds_exists = 0;

	my $esx_server_views = Vim::find_entity_views(view_type => "HostSystem",
							properties => [ 'name',
									'datastore' ] );

	foreach ( @{$esx_server_views || []} ) {
		my $current_esx_name = $_->get_property('name');
		if ($current_esx_name eq $hostname) {
			# Need to create the NAS datastore on this ESX server
			# (OR) Just check if the NAS Datastore exists on this ESX server or not

			print STDERR "Looking at datastores under ESX server $current_esx_name\n";
			my $datastores = $_->datastore;
			foreach ( @{$datastores} ) {
				# Traverse through the Datastores that belong to this ESX server
				my $ds_view = Vim::get_view( mo_ref => $_,
							properties => [ 'name', 'info' ] );
				$ds_name = $ds_view->name;
				if ($ds_name =~ /^$nas_ds_name_prefix/i) {
					# There is already a WIR/IR Datastore created on this ESX Server
					# Now, check if it was created from this same DPU/UEB

					my $remote_host = $ds_view->get_property('info.nas.remoteHost');
					print STDERR "  Remote host is $remote_host\n";
					if (defined $remote_host) {
						# This is a NAS Datastore
						if ($remote_host eq $nas_ds_host) {
							print STDERR " NAS Datastore has already been created from this UEB-RS appliance." .
							" Not creating a new one.";
							$ds_moref = $ds_view->get_property('mo_ref');
							$nas_ds_exists = 1;
							last;
						} else {
							print STDERR " NAS Datastore been created on this ESX server but not from this UEB-RS" .
							" appliance.\n";
						}
					}
				}
			}

			last; # Found what we were looking for. Return.
		}
	} # End forach $esx_server_views


	if ($nas_ds_exists) {
		# Don't create a new NAS Datastore
		return ($nas_ds_exists, $ds_name, $ds_moref);
	} else {
		# Create a new NAS Datastore from this appliance
		return ($nas_ds_exists, 0, 0);
	}
}

sub create_vm_wir {

	my $newvmname      = shift;
	my $pool           = shift;
	my $datastore_name = shift;
	my $hostname       = shift;
	my $disksize       = shift;
	my $floppy         = shift;
	my $switch	   = shift;
	my $efi		   = shift;
	my $os		   = shift;
	my $cpu		   = shift;
	my $boot_disk_num  = shift;
	my $xmldoc         = shift;
	my $xmlroot        = shift;

	my $wir_mode = 1;

	if ( ($os eq "Windows XP") and ( ($cpu eq "Pentium") or ($cpu eq "X86") ) ) {
		# Windows XP 32-bit is not supported
		die "Windows Instant Recovery of Windows XP 32-bit systems is not supported on VMware.\n";
	}

	# $floppy is of the form "<DPU/UEB IP>:<Path to floppy>"
	# Eg: 192.168.134.135:/usr/bp/wvf/12/Static.vfd
	my ($nas_ds_host, $floppy_path) = split(/:/, $floppy);

	my $wir_ds_exists;
	my $nas_ds_name_prefix = "Unitrends_WIR";
	($wir_ds_exists, $nas_ds_name, $nas_ds_moref) = check_nas_ds_exists($hostname, $nas_ds_host, $nas_ds_name_prefix);

	if ($wir_ds_exists) {
		print STDERR "NFS share/WIR Datastore already exists\n";
	} else {
		# Setup NFS to use winbm.iso file to boot from
		my $pid = $$;
		setup_nfs($pid);
		my $nas_ds_path = "/usr/bp/wvf";
		$nas_ds_name = "Unitrends_WIR_$$";
		print STDERR "Name of Datastore is $nas_ds_name\n";
		$nas_ds_moref = create_datastore($hostname, $nas_ds_host, $nas_ds_path, $nas_ds_name);
		print STDERR "MoRef is $nas_ds_moref\n";
	}

	# In some cases (after the NAS Datastore is created), a 'Rename Datastore' command
	# is issued from 'somewhere', which changes the name of the NAS Datastore
	if (!$wir_ds_exists) {
		sleep 20;
	}

	my $vm_view = Vim::get_view(mo_ref => $nas_ds_moref);
	print STDERR "Name of DS is from MoRef is " .$vm_view->name. "\n";
	# A NAS Datastore created with the same mount point on different ESX servers within
	# the same vCenter gets the same name as the one that was created initially on another ESX
	# server within the same vCenter
	$nas_ds_name = $vm_view->name;

	print STDERR "Creating an XML for the VM\n";
	print STDERR "\n\nNew VM name is - $newvmname\n";
	print STDERR "RP name is $pool\n";
	print STDERR "Datastore name is $datastore_name\n";
	print STDERR "Hostname is $hostname\n";
	print STDERR "Disk size is $disksize\n";
	print STDERR "Floppy path is $floppy_path\n";
	print STDERR "Switch is $switch\n";
	print STDERR "EFI is $efi\n";
	print STDERR "Guest OS is $os\n";
	print STDERR "CPU type is $cpu-bit\n";
	print STDERR "Nas DS Name is $nas_ds_name\n";
	print STDERR "Nas DS MoRef is $nas_ds_moref\n";
	print STDERR "Boot disk number is $boot_disk_num\n";
	
	my $xmldoc_vm = VProtect::create_xml_for_vm($newvmname, $pool, $datastore_name, $hostname, $disksize, $floppy_path, $switch, $efi, $os, $cpu, $nas_ds_name, $nas_ds_moref, $boot_disk_num, $xmldoc, $xmlroot);

	my $metadata = "YES";
	my $new_vm_moref = create_vm($xmldoc_vm, $datastore_name, $hostname, $newvmname, $metadata, $pool, $wir_mode);
	print STDERR "Created VM for WIR successfully\n";

	change_boot_order($new_vm_moref, $xmldoc, $xmlroot);

	my $network_changed = 0;
	$network_changed = check_network_setting($new_vm_moref, $hostname, $switch, $xmldoc, $xmlroot);
	if ($network_changed) {
		print STDERR "NIC card setting for this WIR VM was changed\n";
	} else {
		print STDERR "Connected to a Standard switch. So, NIC reconfiguration is not required.\n";
	}

	# TODO When should the NAS DS be removed ?
}

# Sub-routine to reconfigure a VM's Memory and Number of CPUs
sub reconfigure_vm {

	my $vm_uuid = shift;
	my $numcpu = shift;
	my $memory = shift;

	my $spec = VirtualMachineConfigSpec->new(memoryMB => $memory,
						numCPUs => $numcpu);

	my $vm_moref_find_view = Vim::find_entity_view(
					view_type => 'VirtualMachine',
					filter => { 'config.uuid' => $vm_uuid }
					);

	my $task_ref;
	my $retry = 1;
	print STDERR "Before ReconfigVM_Task for changing CPU and memory\n";

	eval {
		$task_ref = $vm_moref_find_view->ReconfigVM_Task(spec => $spec);
	};
	if ($@) {
		if (handle_reconfigvm_error($retry, \$@) == 1) {
			# handle_reconfigvm_error will return 1, when ReconfigVM_Task API fails to
			# reconfigure the VM due to "TaskInProgress" error.
			# Worth to try the reconfigure operation again.
			eval {
				$task_ref = $vm_moref_find_view->ReconfigVM_Task(spec => $spec);
			};
			if ($@) {
				$retry = 0;
				handle_reconfigvm_error($retry, \$@);
			}
		}
	}

	my $msg = "Successfully changed the memory and CPU";
	VProtect::getStatus($task_ref, $msg);

	print STDERR "Reconfigure VM task to change the memory and CPU was successful\n";

}

sub disconnect_cdrom {

	my $vm_moref = shift;

	my $conn = VirtualDeviceConnectInfo->new(
		allowGuestControl => 1,
		connected => 0,
		startConnected => 0
	);

	my $backing = VirtualCdromRemotePassthroughBackingInfo->new(
			deviceName => "",
			exclusive => 0
			);

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);
	my $devices = $vm_view->config->hardware->device;

	my $dc_array_ref = [];

	foreach (@$devices) {
		if ($_->isa('VirtualCdrom')) {
			my %hopt = ( connectable => $conn,
					key => $_->key,
					controllerKey => $_->controllerKey,
					backing => $backing
				);
			my $dev = VirtualCdrom->new(%hopt);

			my $spec = VirtualDeviceConfigSpec->new(
				device => $dev,
				operation => VirtualDeviceConfigSpecOperation->new('edit')
			);

			push @$dc_array_ref, $spec;
		}
	}

	if (scalar @$dc_array_ref > 0) {
		my $change_spec = VirtualMachineConfigSpec->new(
				deviceChange => $dc_array_ref
		);

		my $task_ref;
		my $retry = 1;
		print STDERR "Before ReconfigVM_Task for disconnecting the CD-ROM\n";

		eval {
			$task_ref = $vm_view->ReconfigVM_Task(spec => $change_spec);
		};
		if ($@) {
			if (handle_reconfigvm_error($retry, \$@) == 1) {
				# handle_reconfigvm_error will return 1, when ReconfigVM_Task API fails to
				# reconfigure the VM due to "TaskInProgress" error.
				# Worth to try the reconfigure operation again.
				eval {
					$task_ref = $vm_view->ReconfigVM_Task(spec => $change_spec);
				};
				if ($@) {
					# Die with an appropriate error related to 'ReconfigVM_Task' API
					$retry = 0;
					handle_reconfigvm_error($retry, \$@);
				}
			}
		}

		my $msg = "Successfully disconnected the CD-ROM";
		VProtect::getStatus($task_ref, $msg);
	}

        print STDERR "Reconfigure VM task to disconnect the CD-ROM was successful\n";
	
}

# This sub-routine is called for Audit as well as Verify Mode
# TODO Change the name of the sub-routine
sub audit_vm {

	my $vm_uuid = shift;
	my $numcpu  = shift;
	my $memory  = shift;
	my $xmldoc  = shift;
	my $xmlroot = shift;

	print STDERR "Auditing VM ! \n";
	
	my $vm_moref = VProtect::vim_get_vm_by_uuid($vm_uuid);
	if (!defined $vm_moref) {
		die "VM with UUID ".$vm_uuid." was not found.\n";
	}

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);
	my $firmware = $vm_view->config->firmware;
	print STDERR "The BIOS for this VM is " . $firmware . "\n";

	reconfigure_vm ($vm_uuid, $numcpu, $memory);

	disconnect_network($vm_moref);

	if ($firmware eq 'efi') {
		# Need to disconnect CDROM for EFI, since EFI BIOS looks for a CDROM
		# at boot time, which we don't want for Audit Mode
		disconnect_cdrom($vm_moref);
	}

	power_on_vm($vm_moref);
}

sub reconfig_cd_and_floppy {

	print STDERR "Entering reconfig_cd_and_floppy\n";

	my $vm_moref = shift;
	my $nas_ds_moref_new = shift;
	my $nas_ds_name = shift;

	my $device_array_ref = [];
	my $conn;
	my $backing;
	my %hopt;
	my $dev;
	my $spec;

	$conn = VirtualDeviceConnectInfo->new(
			allowGuestControl => 1,
			connected => 1,
			startConnected => 1
	);

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);
	my $devices = $vm_view->config->hardware->device;

	foreach (@$devices) {
		if ($_->isa('VirtualCdrom')) {
			my $cdrom_filename = "[$nas_ds_name] winbm.iso";
			$backing = VirtualCdromIsoBackingInfo->new(
					datastore => $nas_ds_moref_new,
					fileName => $cdrom_filename
			);

			%hopt = ( connectable => $conn,
					key => $_->key,
					controllerKey => $_->controllerKey,
					backing => $backing
			);
			$dev = VirtualCdrom->new(%hopt);

			$spec = VirtualDeviceConfigSpec->new(
				device => $dev,
				operation => VirtualDeviceConfigSpecOperation->new('edit')
			);

			push @$device_array_ref, $spec;
		}

		if ($_->isa('VirtualFloppy')) {
			print STDERR "Floppy backing type is - ".ref($_->backing)."\n";

			if (ref($_->backing) eq 'VirtualFloppyImageBackingInfo') {
				my $floppy_full_path = $_->backing->fileName;
				my ($ds_name, $floppy_path) = split(/]/, $floppy_full_path);

				if ( ($ds_name or $floppy_path) eq undef) {
					die "Cannot determine the location of the floppy file for " .
					$vm_view->name . "\n";
				}

				my $final_floppy_name = "[$nas_ds_name]$floppy_path";
				$backing = VirtualFloppyImageBackingInfo->new(
						datastore => $nas_ds_moref_new,
						fileName => $final_floppy_name
				);

				%hopt = ( connectable => $conn,
						key => $_->key,
						controllerKey => $_->controllerKey,
						backing => $backing
				);
				$dev = VirtualFloppy->new(%hopt);

				$spec = VirtualDeviceConfigSpec->new(
					device => $dev,
					operation => VirtualDeviceConfigSpecOperation->new('edit')
				);

				push @$device_array_ref, $spec;
			} else {
				die "The floppy type was changed for the VM " . $vm_view->name .
				" Cannot determine the floppy location for WIR purposes.\n";
			}
		}
	}

	if (scalar @$device_array_ref > 0) {
		my $change_spec = VirtualMachineConfigSpec->new(
				deviceChange => $device_array_ref
		);

		my $task_ref;
		my $retry = 1;
		print STDERR "Before ReconfigVM_Task for changing CDROM and Floppy.\n";

		eval {
			$task_ref = $vm_view->ReconfigVM_Task(spec => $change_spec);
		};
		if ($@) {
			if (handle_reconfigvm_error($retry, \$@) == 1) {
				# handle_reconfigvm_error will return 1, when ReconfigVM_Task API fails to
				# reconfigure the VM due to "TaskInProgress" error.
				# Worth to try the reconfigure operation again.
				eval {
					$task_ref = $vm_view->ReconfigVM_Task(spec => $change_spec);
				};
				if ($@) {
					$retry = 0;
					handle_reconfigvm_error($retry, \$@);
				}
			}
		}

		my $msg = "Successfully changed the CDROM and Floppy";
		VProtect::getStatus($task_ref, $msg);

		print STDERR "Reconfigure VM task to change the CDROM and Floppy was successful\n";
	}

	print STDERR "Leaving reconfig_cd_and_floppy\n";

}

sub check_accessibility {

	my $vm_uuid = shift;
	my $vm_moref = shift;

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);

	my $content = Vim::get_service_content();
	my $about = $content->about;
	if ($about->{'apiType'} ne 'VirtualCenter') {
		# If not a vCenter, then, we should still be able to access
		# the NAS share from the appliance
		return;
	}

	my $vm_found = 0;
	my $ds_found = 0;
	my $datastores;
	my $vm_views;

	my $own_ip = `hostname -i`;
	chomp($own_ip);
	print STDERR "Own IP is $own_ip\n";

	my $datacenter_views = VProtect::vim_get_hosts('HostSystem', Opts::get_option('uuid'));

	foreach (@$datacenter_views) {
		print STDERR "Current ESX server UUID is - ".$_->summary->hardware->uuid."\n";
		print STDERR "Current ESX server name is - ".$_->summary->config->name."\n";

		my $esx_name = $_->summary->config->name;
		my $esx_uuid = $_->summary->hardware->uuid;

		$vm_views = VProtect::vim_get_virtual_machines($esx_uuid);

		foreach(@$vm_views) {
			if ($vm_uuid eq $_->config->uuid) {
				# Found the VM that we were looking for
				print STDERR "VM " . $_->config->name . " belongs to ESX server " . $esx_uuid . "\n";
				$vm_found = 1;
				last; # Exit @$vm_views
			}
		}

		if ($vm_found) {

			# Since the VM was found under this ESX server, now check if the
			# Unitrends_WIR_xxx DS is accessible from this ESX server
			$datastores = $_->datastore;
			foreach (@$datastores) {
				my $ds_view = Vim::get_view(mo_ref => $_);
				my $ds_name = $ds_view->get_property('summary')->name;
				print STDERR "Current datastore name is $ds_name.\n";
				if ($ds_name =~ '^Unitrends_WIR') {
					# There is already a WIR Datastore created on this ESX Server
					# Now, check if it was created from this appliance
					my $ds_info = $ds_view->get_property('info');
					my $remoteHost = $ds_info->nas->remoteHost;

					print STDERR "Remote Host is $remoteHost\n";

					if ( $remoteHost eq $own_ip ) {
						# This means that the datastore was created from this same appliance
						# Now check, if it is accessible currently
						$ds_found = 1;  # Unitrends_WIR_xxx NAS datastore exists on this ESX server
						my $ds_accessible = $ds_view->get_property('summary')->accessible;
						print STDERR "Datastore accessible value is - $ds_accessible\n";
						if ($ds_accessible) {
							# Things are good. Return.
							print STDERR "The NAS datastore $ds_name is accessible from $esx_name.\n";
							return;
						} else {
							# This means that the datastore was created from this appliance, but is
							# currently not accessible.
							die "The Unitrends datastore to perform Restores is currently accessible.\n";
						}
					}
				}
			}

			# This means that the WIR VM was found on an ESX server where the
			# NAS datastore was not created. So, create it
			my $pid = $$;
			setup_nfs($pid);
			my $nas_ds_path = "/usr/bp/wvf";
			my $nas_ds_name = "Unitrends_WIR_$pid";
			print STDERR "Name of Datastore is $nas_ds_name\n";
			my $nas_ds_moref_new = create_datastore($esx_name, $own_ip, $nas_ds_path, $nas_ds_name);
			# In some cases, a 'Rename Datastore' command is issued from 'somewhere',
			# which changes the name of the NAS Datastore
			sleep 20;
			print STDERR "MoRef is $nas_ds_moref_new\n";
			my $ds_view = Vim::get_view(mo_ref => $nas_ds_moref_new);
			$nas_ds_name = $ds_view->name;
			print STDERR "NAS datastore name is $nas_ds_name\n";

			reconfig_cd_and_floppy($vm_moref, $nas_ds_moref_new, $nas_ds_name);

			last; # Exit @$datacenter_views
		} # End if($vm_found)
	} # End @$datacenter_views

}

sub restore_vm_wir {

	my $vm_uuid = shift;
	my $numcpu  = shift;
	my $memory  = shift;
	my $xmldoc  = shift;
	my $xmlroot = shift;

	print STDERR "Putting the VM in Restore mode\n";

	my $vm_moref = VProtect::vim_get_vm_by_uuid($vm_uuid);
	if (!defined $vm_moref) {
		die "VM with UUID ".$vm_uuid." was not found.\n";
	}

	# First check if the datastore to access the winbm.iso and floppy is still accessible
	check_accessibility($vm_uuid, $vm_moref);

	reconfigure_vm ($vm_uuid, $numcpu, $memory);

	power_on_vm($vm_moref);

}

=for comment
sub verify_vm {

	my $vm_uuid = shift;
	my $numcpu  = shift;
	my $memory  = shift;
	my $xmldoc  = shift;
	my $xmlroot = shift;

	print STDERR "Verifying VM ! \n";

	my $vm_moref = VProtect::vim_get_vm_by_uuid($vm_uuid);
	if (!defined $vm_moref) {
		die "VM with UUID ".$vm_uuid." was not found.\n";
	}

	my $task_ref = $vm_moref_find_view ->CreateScreenshot_Task();
	my $msg = "Successfully captured screenshot for VM with UUID $vm_uuid !!!";
	my $screenshot_path = &getStatus($task_ref, $msg);

	my $index = rindex($screenshot_path,'/');
	my $screenshot_name = substr($screenshot_path,$index+1);
	$screenshot_path =~ s/^\/vmfs\/volumes\/([a-z0-9]*-[a-z0-9]*)*//g;
	my $screenshot_file = substr($screenshot_path,1,256);

	my $screenshot_dir = "vm_screenshots";
	if( ! -d "$screenshot_dir") {
		mkdir("$screenshot_dir", 0777);
	}

	print STDERR "Downloading \"$screenshot_name\" to \"$screenshot_dir/$screenshot_name\"\n";

	#my $out= `vifs --server "$host" --username "$host_username" --password "$host_password" --get "$datastore_name] $screenshot_file" "$screenshot_dir/$screenshot_name" 2>&1`;
	my $out= `vifs --server "192.168.237.159" --username "admin" --password "Unitrends1!" --get "datastore1] $screenshot_file" "$screenshot_dir/$screenshot_name" 2>&1`;

	print STDERR "Removing screenshot \"$screenshot_name\" from ESX(i) host ...\n";

	#`vifs --server "$host" --username "$host_username" --password "$host_password" --rm "$datastore_name] $screenshot_file" "$screenshot_dir/$screenshot_name" --force 2>&1`;

}
=cut

# This sub-routine assumes that the "audit_vm" sub-routine has been called previously
sub take_screenshot_wir {

	# This functionality is not supported on 7.4
	die "Taking a screenshot is not supported on VMware.\n";

=for comment
	my $vm_uuid = shift;
	my $location = shift;
	my $xmldoc  = shift;
	my $xmlroot = shift;

	print STDERR "Taking screenshot !\n";

	my $vm_moref = VProtect::vim_get_vm_by_uuid($vm_uuid);
	if (!defined $vm_moref) {
		die "VM with UUID ".$vm_uuid." was not found.\n";
	}

	my $vm_moref_find_view = Vim::find_entity_view(
					view_type => 'VirtualMachine',
					filter => { 'config.uuid' => $vm_uuid }
	);

	print STDERR "Power State is - ".$vm_moref_find_view->runtime->powerState->val."\n";
	
	# Before taking a screenshot, make sure that the VM is powered ON
	if ($vm_moref_find_view->runtime->powerState->val ne "poweredOn") {
		die "The VM " . $vm_moref_find_view->name . " is currently not powered ON." .
		" Cannot take a screenshot.\n " ;
	}

	my $task_ref = $vm_moref_find_view ->CreateScreenshot_Task();
	my $msg = "Successfully captured screenshot for VM with UUID $vm_uuid !!!";
	my $screenshot_path = &getStatus($task_ref, $msg);

	# location is of the form '/usr/bp/wvf/<node_no>/<file_name>'

	my @location = split(/\//, $location);
	#print STDERR Dumper(@location);

	print STDERR "Filename is - " . $location[5] . "\n";

	open my $fileHandle, ">>", "$location" or die "Cannot open testfile";
	close $fileHandle;

	# Tried creating a snapshot to see if it generates any activity in the VM
	#		-- Still got a blank screenshot

	# Tried initiating a VMware Tools installer using MountToolsInstaller()
	# to see if it generates any activity in the VM
	#		-- Still got a blank screenshot

	# Tried rebooting the VM to see if there is any fresh activity on the screen
	#		-- Still got a blank screenshot

	my $vmticket = $vm_moref_find_view->AcquireTicket( ticketType => 'mks');
	print STDERR "vmticket is $vmticket\n";
	print STDERR "ref of vmticket is " . ref($vmticket) ."\n\n";

	print STDERR "Config file is - " . $vmticket->cfgFile . "\n";
	print STDERR "Port number is - " . $vmticket->port . "\n";
	print STDERR "Ticket is      - " . $vmticket->ticket . "\n";
	print STDERR "Host is        - " . $vmticket->host . "\n";
	print STDERR "VM name is     - " . $vm_moref_find_view->name . "\n\n";

	eval {
		my $result = $vmticket->sendCAD();
	};
	if ($@) {
		print STDERR Dumper($@);
	}
=cut

}

sub live_vm {

	my $vm_uuid = shift;
	my $numcpu  = shift;
	my $memory  = shift;
	my $xmldoc  = shift;
	my $xmlroot = shift;
	my $warning = "";

	print STDERR "VM going in Live Mode ! \n";

	my $vm_moref = VProtect::vim_get_vm_by_uuid($vm_uuid);
	if (!defined $vm_moref) {
		die "VM with UUID ".$vm_uuid." was not found.\n";
	}

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);
	my $firmware = $vm_view->config->firmware;
	print STDERR "The BIOS for this VM is " . $firmware . "\n";

	if ($firmware eq 'efi') {
		# Need to disconnect CDROM for EFI, since EFI BIOS looks for a CDROM
		# at boot time, which we don't want for Live Mode
		disconnect_cdrom($vm_moref);
	}

	reconfigure_vm ($vm_uuid, $numcpu, $memory);

	connect_network($vm_moref);

	power_on_vm($vm_moref);	

}

sub power_on_vm_wir {

	my $vm_uuid = shift;
	my $numcpu  = shift;
	my $memory  = shift;
	my $xmldoc  = shift;
	my $xmlroot = shift;

	my $vm_moref = VProtect::vim_get_vm_by_uuid($vm_uuid);
	if (!defined $vm_moref) {
		die "VM with UUID ".$vm_uuid." was not found. Unable to power ON\n";
	}

	reconfigure_vm ($vm_uuid, $numcpu, $memory);

	power_on_vm($vm_moref);

}

sub create_snapshot_wir {

	my $vmuuid = shift;
	my $xmldoc  = shift;
	my $xmlroot = shift;
	my $warning = "";

	print STDERR "Creating a snapshot for WIR for VM\n";

	my $vm_moref = VProtect::vim_get_vm_by_uuid($vmuuid);
	if (!defined $vm_moref) {
		die "VM with UUID ".$vmuuid." was not found.\n";
	}

	# Need to call CreateSnapshot on a view, not a moref.
	my $vm_view = Vim::get_view(mo_ref => $vm_moref);

	my $snap_moref;  # New moref of created snapshot.

	my $snap_name = "Unitrends_WIR";
	my $timestamp = strftime("%H:%M %Z %a %b %e %Y", localtime);
	my $description = "Created for Windows Instant Recovery related operations from an Unitrends appliance: " . $timestamp;

	eval {
		# Snap the memory. Hence, memory = 1
		# VMware Tools are not installed on the WIR VM. Hence, quiesce = 0
		$snap_moref = $vm_view->CreateSnapshot(
				name => $snap_name,
				description => $description,
				memory => 1,
				quiesce => 0 );
	};

	if ($@) {
		VProtect::Backup::handle_snapshot_error(0, \$@);
	}

}

# Helper sub-routine for Revert Snapshot operation
sub revert_snap ($$);
sub revert_snap ($$) {

	# Revert to snapshot named $snapshot_name
	my ($snapshot_tree, $snapshot_name) = @_;

	if ($snapshot_tree->childSnapshotList) {
		foreach (@{$snapshot_tree->childSnapshotList}) {
			revert_snap($_, $snapshot_name);
		}
	}

	if ($snapshot_tree->name eq $snapshot_name) {
		my $snap_view = Vim::get_view(mo_ref => $snapshot_tree->snapshot);
		eval {
			$snap_view->RevertToSnapshot();
		};
		if ($@) {
			my $hdr = "Could not revert to Unitrends_WIR snapshot";
			die "$hdr: " . Dumper($@) . "\n";
		}
	}

}

# Revert to snapshot that was created for WIR operations
sub revert_snapshot_wir {

	my $vm_uuid = shift;
	my $xmldoc  = shift;
	my $xmlroot = shift;

	my $vm_moref = VProtect::vim_get_vm_by_uuid($vm_uuid);
	if (!defined $vm_moref) {
		die "VM with UUID ".$vm_uuid." was not found. Unable to revert to snapshot\n";
	}

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);
	my $snapshot_name = "Unitrends_WIR";

	if (defined $vm_view->snapshot) {
		if (defined (@{$vm_view->snapshot->rootSnapshotList})) {
			foreach (@{$vm_view->snapshot->rootSnapshotList}) {
				revert_snap($_, $snapshot_name);
			}
		}
	} else {
		die "No Unitrends Windows Instant Recovery Snapshot exists for this Virtual Machine.\n";
	}

}

# Remove Snapshot that was created for WIR operations
sub remove_snapshot_wir {

	my $vm_uuid = shift;
	my $xmldoc  = shift;
	my $xmlroot = shift;

	my $vm_moref = VProtect::vim_get_vm_by_uuid($vm_uuid);
	if (!defined $vm_moref) {
		die "VM with UUID ".$vm_uuid." was not found. Unable to remove snapshot\n";
	}

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);
	my $ch_present = 0; # Consolidate helper snapshot is present (Only a place holder)
	my $snapshot_name = "Unitrends_WIR";

	if (defined $vm_view->snapshot) {
		if (defined (@{$vm_view->snapshot->rootSnapshotList})) {
			foreach (@{$vm_view->snapshot->rootSnapshotList}) {
				VProtect::Backup::purge_snaps($_, \$ch_present, $snapshot_name);
			}
		}		
	} else {
		die "No Unitrends Windows Instant Recovery Snapshot exists for this Virtual Machine.\n";
	}

}

sub power_off_vm_wir {

	my $vm_uuid = shift;
	my $xmldoc  = shift;
	my $xmlroot = shift;

	my $vm_moref = VProtect::vim_get_vm_by_uuid($vm_uuid);
	if (!defined $vm_moref) {
		die "VM with UUID ".$vm_uuid." was not found. Unable to power OFF\n";
	}

	power_off_vm($vm_moref);

}

sub is_vm_powered_on {

	my $vm_uuid = shift;
	my $xmldoc  = shift;
	my $xmlroot = shift;

	my $vm_moref = VProtect::vim_get_vm_by_uuid($vm_uuid);
	if (!defined $vm_moref) {
		die "VM with UUID ".$vm_uuid." was not found. Unable to check power state\n";
	}

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);

	print STDERR "vm_view->runtime->powerState is ".$vm_view->runtime->powerState->val."\n";
	if ($vm_view->runtime->powerState->val eq "poweredOn" ) {
		print STDERR "VM is powered ON\n";
	} else {
		die "VM is not powered ON\n";
	}
}

sub is_vm_powered_off {

	my $vm_uuid = shift;
	my $xmldoc  = shift;
	my $xmlroot = shift;

	my $vm_moref = VProtect::vim_get_vm_by_uuid($vm_uuid);
	if (!defined $vm_moref) {
		die "VM with UUID ".$vm_uuid." was not found. Unable to check power state\n";
	}

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);

	print STDERR "vm_view->runtime->powerState is ".$vm_view->runtime->powerState->val."\n";
	if ($vm_view->runtime->powerState->val eq "poweredOff" ) {
		print STDERR "VM is powered OFF\n";
	} else {
		die "VM is not powered OFF\n";
	}

}

sub delete_vm {

	my $vm_uuid = shift;
	my $xmldoc  = shift;
	my $xmlroot = shift;

	print STDERR "Caution: Deleting the VM ! \n";

	my $vm_moref = VProtect::vim_get_vm_by_uuid($vm_uuid);
	if (!defined $vm_moref) {
		# *********** CAUTION ***********
		# This error is consumed as it is in the layer above. 
		# DO NOT CHANGE THIS ERROR
		# If required to change, then, make sure the same change 
		# goes in the above layer
		die "VM with UUID ".$vm_uuid." was not found while deleting the VM.\n";
	}

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);

	# Could be the case that the VM is already powered ON
	# So, first power OFF the VM
	if ($vm_view->runtime->powerState->val eq "poweredOn") {
		power_off_vm($vm_moref);
	}

	my $task_ref;
	eval {
		my $task = $vm_view->Destroy();
	};
	if ($@) {
		my $hdr = 'Failed to Delete the VM: ';
		if (ref($@) eq 'SoapFault') {
			if (ref($@->detail) eq 'RuntimeFault') {
				die $hdr . "RuntimeFault.\n" . Dumper($@) . "\n";
			}
			if (ref($@->detail) eq 'VimFault') {
				die $hdr . "VimFault.\n" . Dumper($@) . "\n";
			}
		}

		# Failed to identify the fault.
		die "Failed to delete the VM.\n" . Dumper($@);
	}

}


1;
